r"""## Quark and Gluon Jets

Four datasets of quark and gluon jets, each having two million total jets, have
been generated with [Pythia](http://home.thep.lu.se/~torbjorn/Pythia.html) and
[Herwig](https://herwig.hepforge.org/) and are accessible through this
submodule of EnergyFlow. The four datasets are:

- Pythia 8.226 quark (uds) and gluon jets.
- Pythia 8.235 quark (udscb) and gluon jets.
- Herwig 7.1.4 quark (uds) and gluon jets.
- Herwig 7.1.4 quark (udscb) and gluon jets

To avoid downloading unnecessary samples, the datasets are contained in twenty
files with 100k jets each, and only the required files are downloaded. These
are based on the samples used in 
[1810.05165](https://arxiv.org/abs/1810.05165). Splitting the data into 
1.6M/200k/200k train/validation/test sets is recommended for standardized
comparisons.

Each dataset consists of two components:

- `X` : a three-dimensional numpy array of the jets with shape 
`(num_data,max_num_particles,4)`.
- `y` : a numpy array of quark/gluon jet labels (quark=`1` and gluon=`0`).

The jets are padded with zero-particles in order to make a contiguous array.
The particles are given as `(pt,y,phi,pid)` values, where `pid` is the
particle's [PDG id](http://pdg.lbl.gov/2018/reviews/rpp2018-rev-monte
-carlo-numbering.pdf). Quark jets either include or exclude $c$ and $b$
quarks depending on the `with_bc` argument.

The samples are generated from $q\bar q\to Z(\to\nu\bar\nu)+g$ and
$qg\to Z(\to\nu\bar\nu)+(uds[cb])$ processes in $pp$ collisions at
$\sqrt{s}=14$ TeV. Hadronization and multiple parton interactions (i.e.
underlying event) are turned on and the default tunings and shower parameters
are used. Final state non-neutrino particles are clustered into $R=0.4$
anti-$k_T$ jets using FastJet 3.3.0. Jets with transverse momentum
$p_T\in[500,550]$ GeV and rapidity $|y|<1.7$ are kept. Particles are ensured
have to $\phi$ values within $\pi$ of the jet (i.e. no $\phi$-periodicity 
issues). No detector simulation is performed.

The samples are also hosted on Zenodo and we ask that you cite them
appropriately if they are useful to your research. For BibTex entries,
see the [FAQs](/faqs/#how-do-i-cite-the-energyflow-package).

[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.3164691.svg)](https://doi.org/10.5281/zenodo.3164691) - Pythia samples
<br>
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.3066475.svg)](https://doi.org/10.5281/zenodo.3066475) - Herwig samples
"""
from __future__ import absolute_import, division, print_function

import warnings

import numpy as np

from energyflow.utils.data_utils import _get_filepath, _pad_events_axis1

__all__ = ['load']

NUM_PER_FILE = 100000
MAX_NUM_FILES = 20
URLS = {
    'pythia': {
        'nobc': {
            'dropbox': [
                'https://www.dropbox.com/s/fclsl7pukcpobsb/QG_jets.npz?dl=1',
                'https://www.dropbox.com/s/ztzd1a6lkmgovuy/QG_jets_1.npz?dl=1',
                'https://www.dropbox.com/s/jzgc9e786tbk1m5/QG_jets_2.npz?dl=1',
                'https://www.dropbox.com/s/tiwz2ck3wnzvlcr/QG_jets_3.npz?dl=1',
                'https://www.dropbox.com/s/3miwek1n0brbd2i/QG_jets_4.npz?dl=1',
                'https://www.dropbox.com/s/tsq80wc6ngen9kn/QG_jets_5.npz?dl=1',
                'https://www.dropbox.com/s/5oba2h15ufa57ie/QG_jets_6.npz?dl=1',
                'https://www.dropbox.com/s/npl6b2rts82r1ya/QG_jets_7.npz?dl=1',
                'https://www.dropbox.com/s/7pldxfqdb4n0kaw/QG_jets_8.npz?dl=1',
                'https://www.dropbox.com/s/isw4clv7n370nfb/QG_jets_9.npz?dl=1',
                'https://www.dropbox.com/s/prw7myb889v2y12/QG_jets_10.npz?dl=1',
                'https://www.dropbox.com/s/10r4ydro3e6nsmc/QG_jets_11.npz?dl=1',
                'https://www.dropbox.com/s/42p10sv9jedmtn0/QG_jets_12.npz?dl=1',
                'https://www.dropbox.com/s/crqdeg4arjti7cy/QG_jets_13.npz?dl=1',
                'https://www.dropbox.com/s/1e7ss2quxhkbhwy/QG_jets_14.npz?dl=1',
                'https://www.dropbox.com/s/psje9feje43buc7/QG_jets_15.npz?dl=1',
                'https://www.dropbox.com/s/8qw5bcswgrr9fl1/QG_jets_16.npz?dl=1',
                'https://www.dropbox.com/s/gcdp98bgupfk05x/QG_jets_17.npz?dl=1',
                'https://www.dropbox.com/s/jvgt17z1ufxz1ly/QG_jets_18.npz?dl=1',
                'https://www.dropbox.com/s/gbbfvy2e0slmm8v/QG_jets_19.npz?dl=1',
            ],
            'zenodo': [
                'https://zenodo.org/record/3164691/files/QG_jets.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_1.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_2.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_3.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_4.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_5.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_6.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_7.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_8.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_9.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_10.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_11.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_12.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_13.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_14.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_15.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_16.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_17.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_18.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_19.npz?download=1',
            ],
        },

        'bc': {
            'dropbox': [
                'https://www.dropbox.com/s/hlu497verxb9f4x/QG_jets_withbc_0.npz?dl=1',
                'https://www.dropbox.com/s/fi3knsjwg5dvcu6/QG_jets_withbc_1.npz?dl=1',
                'https://www.dropbox.com/s/cooz6qysvnfsqmr/QG_jets_withbc_2.npz?dl=1',
                'https://www.dropbox.com/s/ej7xteeoyc7meau/QG_jets_withbc_3.npz?dl=1',
                'https://www.dropbox.com/s/j2z30kh5u7t3ppb/QG_jets_withbc_4.npz?dl=1',
                'https://www.dropbox.com/s/d94krcfcn6ca98y/QG_jets_withbc_5.npz?dl=1',
                'https://www.dropbox.com/s/b5rfd0z3na09l99/QG_jets_withbc_6.npz?dl=1',
                'https://www.dropbox.com/s/02gkrs0pbpzxwn2/QG_jets_withbc_7.npz?dl=1',
                'https://www.dropbox.com/s/dlvquskq4fn3oy7/QG_jets_withbc_8.npz?dl=1',
                'https://www.dropbox.com/s/5yny7e4l8bu0ps2/QG_jets_withbc_9.npz?dl=1',
                'https://www.dropbox.com/s/93wu2rnnnf7og9u/QG_jets_withbc_10.npz?dl=1',
                'https://www.dropbox.com/s/1p2whcbhc19rusk/QG_jets_withbc_11.npz?dl=1',
                'https://www.dropbox.com/s/o35w4ds3d0jfkl2/QG_jets_withbc_12.npz?dl=1',
                'https://www.dropbox.com/s/shjbg6mluivyyry/QG_jets_withbc_13.npz?dl=1',
                'https://www.dropbox.com/s/k3phvslpc85qudk/QG_jets_withbc_14.npz?dl=1',
                'https://www.dropbox.com/s/vyif9her9nwstx9/QG_jets_withbc_15.npz?dl=1',
                'https://www.dropbox.com/s/jw6e31c6dmhpk4t/QG_jets_withbc_16.npz?dl=1',
                'https://www.dropbox.com/s/lgxce8v7widxzju/QG_jets_withbc_17.npz?dl=1',
                'https://www.dropbox.com/s/bj43a5a8z3nsb4n/QG_jets_withbc_18.npz?dl=1',
                'https://www.dropbox.com/s/jal2p6o85bnj33d/QG_jets_withbc_19.npz?dl=1',
            ],
            'zenodo': [
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_0.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_1.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_2.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_3.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_3.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_4.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_5.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_6.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_7.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_8.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_9.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_10.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_12.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_13.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_14.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_15.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_16.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_17.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_18.npz?download=1',
                'https://zenodo.org/record/3164691/files/QG_jets_withbc_19.npz?download=1',
            ],
        },
    },

    'herwig': {
        'nobc': {
            'dropbox': [
                'https://www.dropbox.com/s/xizexr2tjq2bm59/QG_jets_herwig_0.npz?dl=1',
                'https://www.dropbox.com/s/ym675q2ui3ik3n9/QG_jets_herwig_1.npz?dl=1',
                'https://www.dropbox.com/s/qic6ejl27y6vpqj/QG_jets_herwig_2.npz?dl=1',
                'https://www.dropbox.com/s/ea5a9wruo7sf3zy/QG_jets_herwig_3.npz?dl=1',
                'https://www.dropbox.com/s/5iz5q2pjcys74tb/QG_jets_herwig_4.npz?dl=1',
                'https://www.dropbox.com/s/6zha7fka0dl7t30/QG_jets_herwig_5.npz?dl=1',
                'https://www.dropbox.com/s/vljp5nhoocv2zmf/QG_jets_herwig_6.npz?dl=1',
                'https://www.dropbox.com/s/vzzl5yv9esro811/QG_jets_herwig_7.npz?dl=1',
                'https://www.dropbox.com/s/74u8y4afe1jqiyw/QG_jets_herwig_8.npz?dl=1',
                'https://www.dropbox.com/s/ra7hdq23qy7lgia/QG_jets_herwig_9.npz?dl=1',
                'https://www.dropbox.com/s/plhupkzt3ap2v6i/QG_jets_herwig_10.npz?dl=1',
                'https://www.dropbox.com/s/jy76a7tk1p7b5mq/QG_jets_herwig_11.npz?dl=1',
                'https://www.dropbox.com/s/cd4bqzk1xhg92tp/QG_jets_herwig_12.npz?dl=1',
                'https://www.dropbox.com/s/5g5rbyowni149y4/QG_jets_herwig_13.npz?dl=1',
                'https://www.dropbox.com/s/uxcgkrz4jhwdnya/QG_jets_herwig_14.npz?dl=1',
                'https://www.dropbox.com/s/brgeiph0rhooffx/QG_jets_herwig_15.npz?dl=1',
                'https://www.dropbox.com/s/jvcw8th5t6ngsk1/QG_jets_herwig_16.npz?dl=1',
                'https://www.dropbox.com/s/hlgksqbpuw3wuo3/QG_jets_herwig_17.npz?dl=1',
                'https://www.dropbox.com/s/yjvnt2z0h1zhvns/QG_jets_herwig_18.npz?dl=1',
                'https://www.dropbox.com/s/8qzs744mx383n2i/QG_jets_herwig_19.npz?dl=1',
            ],
            'zenodo': [
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_0.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_1.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_2.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_3.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_4.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_5.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_6.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_7.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_8.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_9.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_10.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_11.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_12.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_13.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_14.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_15.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_16.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_17.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_18.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_19.npz?download=1',
            ],
        },

        'bc': {
            'dropbox': [
                'https://www.dropbox.com/s/qv5t4171ez82kqr/QG_jets_herwig_withbc_0.npz?dl=1',
                'https://www.dropbox.com/s/mae1hmudq8v0tqr/QG_jets_herwig_withbc_1.npz?dl=1',
                'https://www.dropbox.com/s/6yc14771808mf0w/QG_jets_herwig_withbc_2.npz?dl=1',
                'https://www.dropbox.com/s/ihuffb5nzblw2mr/QG_jets_herwig_withbc_3.npz?dl=1',
                'https://www.dropbox.com/s/nygld5xtxmg7id2/QG_jets_herwig_withbc_4.npz?dl=1',
                'https://www.dropbox.com/s/zn76rajxowk91hn/QG_jets_herwig_withbc_5.npz?dl=1',
                'https://www.dropbox.com/s/uajiizu1k5d24x0/QG_jets_herwig_withbc_6.npz?dl=1',
                'https://www.dropbox.com/s/xcw7nfkr4r7mglf/QG_jets_herwig_withbc_7.npz?dl=1',
                'https://www.dropbox.com/s/hlvgl69hig6nepp/QG_jets_herwig_withbc_8.npz?dl=1',
                'https://www.dropbox.com/s/3cbtd73z0mdop7l/QG_jets_herwig_withbc_9.npz?dl=1',
                'https://www.dropbox.com/s/zadw2vjo71mmfkf/QG_jets_herwig_withbc_10.npz?dl=1',
                'https://www.dropbox.com/s/xivt0q49k0vccmy/QG_jets_herwig_withbc_11.npz?dl=1',
                'https://www.dropbox.com/s/ft0z5eagni71c4v/QG_jets_herwig_withbc_12.npz?dl=1',
                'https://www.dropbox.com/s/4wsui0wc0zueq3l/QG_jets_herwig_withbc_13.npz?dl=1',
                'https://www.dropbox.com/s/73kkum4kfm9jxlk/QG_jets_herwig_withbc_14.npz?dl=1',
                'https://www.dropbox.com/s/i4tflx17r1prr5u/QG_jets_herwig_withbc_15.npz?dl=1',
                'https://www.dropbox.com/s/m0xnoauoghg29zr/QG_jets_herwig_withbc_16.npz?dl=1',
                'https://www.dropbox.com/s/dtgyyflmxa86l8o/QG_jets_herwig_withbc_17.npz?dl=1',
                'https://www.dropbox.com/s/nsm8hj1lolz8qk5/QG_jets_herwig_withbc_18.npz?dl=1',
                'https://www.dropbox.com/s/t2vgtj47jy4o1di/QG_jets_herwig_withbc_19.npz?dl=1',
            ],
            'zenodo': [
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_0.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_1.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_2.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_3.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_4.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_5.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_6.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_7.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_8.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_9.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_10.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_11.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_12.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_13.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_14.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_15.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_16.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_17.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_18.npz?download=1',
                'https://zenodo.org/record/3066475/files/QG_jets_herwig_withbc_19.npz?download=1',
            ],
        },
    },
}

HASHES = {
    'pythia': {
        'nobc': {
            'sha256': [
                '3f27a02eab06e8b83ccc9d25638021e6e24c9361341730961f9d560dee12c257',
                '648e49cd59b5353e0064e7b1a3388d9c2f4a454d3ca67afaa8d0344c836ecb35',
                '09f7b16fa7edb312c0f652bb8504de45f082c4193df65204d693155017272fe9',
                '7dc9a50bb38e9f6fc1f11db18f9bd04f72823c944851746b848dee0bba808537',
                '3e6217aad8e0502f5ce3b6371c61396dfc48a6cf4f26ee377cc7b991b1d2b543',
                'b5b7d742b2599bcbe1d7a639895bca64c28da513dc3620b0e5bbb5801f8c88fd',
                '7d31bc48c15983401e0dbe8fd5ee938c3809d9ee3c909f4adab6daf8b73c14f1',
                'cec0d7b2afa9d955543c597f9b7f3b3767812a68b2401ec870caf3a2ceb98401',
                'e984620f57abe06fc5d0b063f9f84ba54bd3e8c295d2b2419a7b1c6175079ed4',
                '6e3b69196995d6eb3b8e7af874e2b9f93d904624f7a7a73b8ff39f151e3bd189',
                'fa3d386f230b806058ff17e5bd77326ff4bf01d72aa5eb3325c1df2a8825927c',
                'acd49ab7bea8f72ecf699a9a898bccacc8730474259d68406656a5a43d407fb0',
                '2edd55b8bc30c686a0637855e1ba068586eb97041e8114d5540d96db2a7a2e17',
                '7276a8a0e573f9795a47f9d5addc10d2af903c2a0ffa5c848a720ccae93daa90',
                '2068ecfa912e94cd3ce7273b7c77af0bbd5ec57940997e7483b56f03434a6869',
                '41a732ce6321dd593214225b03fb87329607ccae768c705e3896ffecc28bfcca',
                '9d68caeb18f3ccf127b9032f52e63ee011c4381293a3a503f894e5c0741ae215',
                '086053ca611bb04d97fa0b6509b4ffb6955421b067c7b277498f0e5188879331',
                'cdc595f5fedef7db9411a9f93f2786f110073b4d17a523700f625846588b1e44',
                'd07781139320ae134ce4824bc0cefa43fd5003cd97cdf3aed90d4fb12fad8a1d',
            ],
            'md5': [
                'f5d052f10a79c6e8b9382637aca0ef52',
                'f6a1081c76a47386bc11abcf0e499552',
                '2628367c57ba598f4473c870d1381041',
                'dd3ad998b0a1bd9acea2ecf029a8a921',
                'a56d6bb98361b55382aa8c06225e05d8',
                '266c688e9e6ff1cd20840692d45eaaf8',
                '95a9f7e555fb7b1073967056b9030b11',
                '4ae72aaabe121bd489532c99a6bdde95',
                'a2b80bd4199468fde4f302d346a8c9d8',
                '1157cbace488c70c9dcfc250f3345b06',
                '4b424b553e1e7f852e47ea9904bc2dcf',
                'ccd29c9d1abb34dd7cfb48cfc57a9695',
                '1ed1f6f19fb8439c9811dced41d5127d',
                'af45818c361e11ca9b3adaba30db06ad',
                '488ced3ea409d7e2b196da67f7d182ec',
                'c5e083019de6cd6a0ef12bcec1ea566b',
                '48605d55edff665f0c7d2f800b5a622e',
                '8fd47760957b5fd9adec9048b50cd1a9',
                'd43d611484b55391e891ba31c605f792',
                '6753508e34014cc69714a01fca20ec38',
            ],
        },

        'bc': {
            'sha256': [
                '27978b5dfe38f860f9899a4213f115579766ece0f6b3cd1cc043f57483521f9c',
                'bec3a147167ac19f243d74c5c47097a716cee6f6af4edc16fd0b50003ec48bd7',
                '878e415001682fda5493f15f2aaa29bce3d60b5dd882f85d85f66f2e8c5ddf9d',
                'e6f48b8fa5dfb3fa914db5dbcabe6d972d803571aa5586babececa86007d0064',
                'e11f885b97b7e859792b3f0a748f15e89c8e8788d68d1256931c948b745945e1',
                '9598efe81b4d1f5f56049508da23957fd5c90590a242f7e58255aaa44d23c192',
                '1de06fccea886445cc4250aba8e3c10990ccb24d4ae6416c6f1f40811117c13e',
                'df6d44f5c37e5c1bf6a8c7cbfb9633413348c908b5f0c657ec32e8dde781ca95',
                'fa6a01f90cdf3394b77bdec7e1931b2dfa4d6670ad8fbc0266eb14c352456e93',
                'e8c7545edb1bc52a0ea0b6cf8541a3ab825970c337458e5c43402c13939e949d',
                '147a6f80eb191577092c3bc404c90ab5538e2887b23f27d6e91e3643c5a18119',
                'c1068b3ac7d0f94ec928538a451d9e20feb7b0281deb083661f8fc6bffa7c4f1',
                '78545b1517f099b0001f6dafe046a9d39a93d13e29db812f6fa415301bfa590d',
                '8d90ffc18358dfb234359804da6a37bb188fa45c9e24468b0cc91f24ba6e0a1d',
                '4ccdf00ef7948721bfe41cd61cb3eac2025e4869b226c9f22524cf2adb9ed2a0',
                'ca09495a8b5b27435d8523aaca6f3af6795a9a61c17d758be2c1c5560e5423f1',
                '00318c05ef530f7b2e867f4b1b9900e9535add46455f6fc39ad4777ff9c71e00',
                'a23c05633f18b1a047102943f550c8bb5cfc62135a2d543ccc7cdef333996f5c',
                '0883dcd7feebef2f6419ca084b9ecc0c28c5ada68eb56f6062a943d3fe4bad81',
                'e60a6c2bac382f51147b8587119b589b134f39724fec598e28193c9b03f70fa5',
            ],
            'md5': [
                'e9ac4044a07f56a919a96e2b30c15fed',
                '5e0c0a08c5de47b190b514ce49ff4e94',
                '7c4209b14f778bb6c8cb6833a3d47854',
                'cabbd75d4313e07bf8cd8e3479e06c18',
                '87e793e74e5e665e40c1ece764952934',
                'fd0ff359e1e64023b1c1f05e854c7180',
                'fed4dfe45618598c853f8d9d24a40afd',
                '4220bee4b081850e41b3462c06538bda',
                'a7d76ec2ab2ef6d5777de5574e289c26',
                '40b9d803d1579ef77d6eda733e385e22',
                'bd5871239ddadcf080e62ffea07bb122',
                'd7731f5abdf7d3e0c17eca846c25eff3',
                '68514a76032dddb3b9cfd38654a4d433',
                'd3dca7cb617c66e6f58ea248a10b5ccb',
                '625a2d19e9b3ac6907362be3cefc404c',
                'dae5923aba89ad393e1c59c63f2552e2',
                'c611ace22f23518ae20d9e828fb5d0bc',
                '821cff3746d2fa07ad1b9feb056dd88b',
                '5c4775a99d18ac713360d9bbc43bfb43',
                '5fd9bcfa8baafb6b3f6efb5114420976',
            ],
        },
    },

    'herwig': {
        'nobc': {
            'sha256': [
                '0527349778c0ab2f7da268975fb9e7c0705c88d60f2c478401d941b9913f4d44',
                '65ef3b4cced4e2618c2bf8f3c66ef707dbd7a9740825f93549732d64e60d7ea8',
                'f13dab1937e40d0c05b97e9813fb4dda5156a8f6b4e41a89cc13821d02a60f58',
                '7b55e26262f2c156b15014b796d0a7e7a5254a982170f45cf2d9857b1f23b5f7',
                '3a5006da4a05192636a74fc818256fce215970c626719738cae9f82e3f068646',
                '2601564aee41aa5851392d6b3d12021f978fa17199e42dde004e35d1175055ea',
                '2c1fc34e99816a0bb5a84f68fa42f6314252521f6b32384a118cdec872ea97a1',
                '4b05f17acb046ad50232987003b89a91800cc713eefd81142ffeb42259369fb2',
                '150cbe132a2ee3178ba3a93a6b1733b3498b728db91f572295b6213d287ec1f7',
                '7d74c90843c751ade4cac47f6c2505da8bcbaf8645bc3f9870bdca481ff805fd',
                'e2b9072da8436618c602fbcf2409fe9be9a46dea7cff1fcc36f1ba8fefa6842d',
                'c69f499b7ea09029da7e78dcc527feca6b1680685e3c9a481db292d5518e3f1c',
                'db25d85d3a35978c607f9b5b0b52f4140c984eb5a5ab236cbf3e6eb34ead761c',
                '9a51ddd383e32154fc504ddcb138e54f0f1bd35079fe5cfa9139839c229cd78e',
                'be0fa462ea907d36972c8573b9a2f6bcdf5cf66648fa397739d12ecb677948e5',
                '7b17400c6867243e8137bd97e0f9743682a5d8c772685a6654f42f1fa2731960',
                '4c484f6508180c0e4e4a5c90b37d1b15cc67afaa3c5998306e8e633848ce6dfc',
                'd1f9baf3a3a148080d1735130f6b18f0f598991a8d886eff3c427b2e2265fce1',
                'bb70219d78e1d92091efacbf933da632670c8318d00777e4144d9a0c782e5749',
                '94975b3d999868485780d2d9e4330273aa8f0db4b9a7f6094d360f637659a264',
            ],
            'md5': [
                'a9de310c35c5a83ea592ef93070ff2f3',
                'd6ff8cc5c6192309fba915114fdc8358',
                '625bc4a0619b5b2551e273be493c6092',
                '821b293d4e68db8b2bd40a1732d1d865',
                '415dded70fca2ae5e555cdee776724d8',
                '242e23df1b837b9afac880383157a161',
                '068eb955146f773c1b5815dd3424c434',
                'e0212b768f57344ae60df7783ba5ba25',
                'd1a082794c84d2b0cc159034cf4d44b6',
                'f2e1c99033a2ff7d97d9968d394333ba',
                '5eab363df8bdff106f53858e60fe7ed1',
                '2b365047d797207009e3b40f3ec71669',
                '02317fda982357aa3874fd8c6c0e5863',
                'a102b2056cc08e7c7a9312461151e749',
                '628acd4a4e7b39d8b8e3675f4d91a3d0',
                'f19512ae26b5a54930bb57d7b3ab8672',
                'c71796ceae838945710c53ce95a33297',
                'b9e26d977f8b0e638f5db42cd1c3bcee',
                '6339009c2f5e06ac3c7b2ddade3b3e68',
                'efdec3aa9194e835f2773be6fb424054',
            ],
        },

        'bc': {
            'sha256': [
                '173b2ee0c5466772997d6b9f6a8ce25531ae3666ee17e73df1807a707aefbb17',
                'd63f272b7a5be9b75ba26082eb76107e882b756a2285cab0c16ed69d77c16366',
                '59369e725de3f688b231993bc7fbca45c2c1cc1da252aee5160298e83ce303cf',
                'bf5f2e8d6ce306796dc4e3ce9e6a88faa6ddb9b482b714060a11aa257b0fe1e6',
                'db5f5dd682f6f48e1b4900e2f366eddbb4089cb14f745cd2fcc79b23ff9f8104',
                '03833b850fa2c1ad050c753b7be9e08086bbb2b55ae41dfc1921c617cc85a622',
                'a7f94530366d886ab16adbb786671782dc4001e76ffb87937e65852449fe4f9c',
                '0c1f326807a5d5d57398aec6c5816d7c9dc644fd80e56af876fa011b139cb163',
                '1f272ba63ccf9d0af72add26075e9fc6c57e4bb954a2a31a542c06d328061742',
                'd36cd5bbaf84a4b01faa807489133ad5ca283f64f1602e724684f1e1c2996be6',
                'a4e735dcf69de0e635974a5eafb14e7cfa894e2db29a342f04be8660dc8f190f',
                'c3b293be7c5cc45f65a94835d4f8c6920abe9b5c8f4baa3f10afe1dfd0112af0',
                '2a1b54e081692eae967f117f6e8867a46f5d7d7a9d7c9353342d3cb72414d62c',
                '6a7e846baccc8076563ed6bcb5349c2f07c4802ffeefed7fe3350680d49fc9a9',
                'fb203a1273e9ecf93176cb63123f4a5ae39d803322bbd666679651715e1c8617',
                'b9493efda0e3e1c4fdcb684486b1157d73a9ca9f29142125cded06d5a85b5df4',
                '1a8e6fcefb81805d7f22bffa5b45a71d9a0fcb97a490d1866865ea7196d94443',
                '2e961954b1ca4642cc2434789e61cd0e2eb8f18f8e77b120950de1c81ed94a15',
                '34129cd813aa54539837ad57e484206f20f1da0330ae5fab6b378707ea41ca25',
                '3895ad282094f3c248bdee5ff1c57247476f03accafd573b60e0ac0d0463fe0b',
            ],
            'md5': [
                '2acf8751843b18d97fc9d2c5cc1ada6e',
                'd900130f470a16edf95faef844169c98',
                '1ff02a8f99e16645ce1aeea262d48e08',
                'cd8c63195966bec92846fd085e696a24',
                '4d1d284cadc6a2c682f7673611c4d583',
                'd6dccec5c6a7c80d967e2cc2bf5955f6',
                'e4bc0643113a41820895a8911c8b54f8',
                '814d109e673c78fd7f7c15143587c78b',
                '252a31659ebacb17ebd1a41250fa8546',
                'adb64fcc128744fbe8944b865a92a6e6',
                'e7c151d8af531840823c53ac67518e2e',
                'e4535cb57361eb31b51542ced6604626',
                'a752c262f3e0ac1eda48d49496eaf46c',
                'ad486b64b76ba73e93ee0ef1aaf2b3ba',
                '0019446390bd46ed7f8679da57c5ced0',
                '8942d2900be5d924ba5fe859daea6947',
                'c9a01fb2d9bd6a4f18ebc721e2d91b1c',
                'c9d31a2b8baeb44d9b1f821b023669b1',
                '253e7ef9d9edf37306f7d2b769377a95',
                '185902f2aba6c7c1b79dea24d9997146',
            ],
        },
    },
}

GENERATORS = frozenset(URLS.keys())
SOURCES = ['dropbox', 'zenodo']

def load(num_data=100000, generator='pythia', pad=True, with_bc=False, cache_dir='~/.energyflow'):
    """Loads samples from the dataset (which in total is contained in twenty 
    files). Any file that is needed that has not been cached will be 
    automatically downloaded. Downloading a file causes it to be cached for
    later use. Basic checksums are performed.

    **Arguments**

    - **num_data** : _int_
        - The number of events to return. A value of `-1` means read in all
        events.
    - **generator** : _str_
        - Specifies which Monte Carlo generator the events should come from.
        Currently, the options are `'pythia'` and `'herwig'`.
    - **pad** : _bool_
        - Whether to pad the events with zeros to make them the same length.
        Note that if set to `False`, the returned `X` array will be an object
        array and not a 3-d array of floats.
    - **with_bc** : _bool_
        - Whether to include jets coming from bottom or charm quarks. Changing
        this flag does not mask out these jets but rather accesses an entirely
        different dataset. The datasets with and without b and c quarks should
        not be combined.
    - **cache_dir** : _str_
        - The directory where to store/look for the files. Note that 
        `'datasets'` is automatically appended to the end of this path.

    **Returns**

    - _3-d numpy.ndarray_, _1-d numpy.ndarray_
        - The `X` and `y` components of the dataset as specified above. If
        `pad` is `False` then these will be object arrays holding the events,
        each of which is a 2-d ndarray.
    """

    # check for valid options
    if generator not in GENERATORS:
        raise ValueError("'generator' must be in " + str(GENERATORS))

    # get number of files we need
    num_files = int(np.ceil(num_data/NUM_PER_FILE)) if num_data > -1 else MAX_NUM_FILES
    if num_files > MAX_NUM_FILES:
        warnings.warn('More data requested than available. Providing the full dataset.')
        num_files = MAX_NUM_FILES
        num_data = -1

    # index into global variables
    bc = 'bc' if with_bc else 'nobc'
    urls = URLS[generator][bc]
    hashes = HASHES[generator][bc]

    # obtain files
    Xs, ys = [], []
    for i in range(num_files):
        for j,source in enumerate(SOURCES):
            try:
                url = urls[source][i]
                filename = url.split('/')[-1].split('?')[0]

                fpath = _get_filepath(filename, url, cache_dir, file_hash=hashes['sha256'][i])

                # we succeeded, so don't continue trying to download this file
                break

            except Exception as e:
                print(str(e))

                # if this was our last source, raise an error
                if j == len(SOURCES) - 1:
                    m = 'Failed to download {} from any source.'.format(filename)
                    raise RuntimeError(m)

                # otherwise indicate we're trying again
                else:
                    print("Failed to download {} from source '{}', trying next source...".format(filename, source))

        # load file and append arrays
        f = np.load(fpath)
        Xs.append(f['X'])
        ys.append(f['y'])
        f.close()

    # get X array
    if pad:
        max_len_axis1 = max([X.shape[1] for X in Xs])
        X = np.vstack([_pad_events_axis1(x, max_len_axis1) for x in Xs])
    else:
        X = np.asarray([x[x[:,0]>0] for X in Xs for x in X])

    # get y array
    y = np.concatenate(ys)

    # chop down to specified amount of data
    if num_data > -1:
        X, y = X[:num_data], y[:num_data]

    return X, y
    


from __future__ import absolute_import, division, print_function

from keras.layers import Dense, Dropout
from keras.models import Sequential
from keras.regularizers import l2

from energyflow.archs.archbase import NNBase
from energyflow.utils import iter_or_rep

__all__ = ['DNN']

###############################################################################
# DNN
###############################################################################

class DNN(NNBase):

    """Dense Neural Network architecture."""

    # DNN(*args, **kwargs)
    def _process_hps(self):
        """See [`ArchBase`](#archbase) for how to pass in hyperparameters as
        well as defaults common to all EnergyFlow neural network models.

        **Required DNN Hyperparameters**

        - **input_dim** : _int_
            - The number of inputs to the model.
        - **dense_sizes** : {_tuple_, _list_} of _int_
            - The number of nodes in the dense layers of the model.

        **Default DNN Hyperparameters**

        - **acts**=`'relu'` : {_tuple_, _list_} of _str_ or Keras activation
            - Activation functions(s) for the dense layers. A single string or
            activation layer will apply the same activation to all dense layers.
            Keras advanced activation layers are also accepted, either as
            strings (which use the default arguments) or as Keras `Layer` 
            instances. If passing a single `Layer` instance, be aware that this
            layer will be used for all activations and may introduce weight 
            sharing (such as with `PReLU`); it is recommended in this case to 
            pass as many activations as there are layers in the model.See the
            [Keras activations docs](https://keras.io/activations/) for more 
            detail.
        - **k_inits**=`'he_uniform'` : {_tuple_, _list_} of _str_ or Keras 
        initializer
            - Kernel initializers for the dense layers. A single string 
            will apply the same initializer to all layers. See the
            [Keras initializer docs](https://keras.io/initializers/) for 
            more detail.
        - **dropouts**=`0` : {_tuple_, _list_} of _float_
            - Dropout rates for the dense layers. A single float will
            apply the same dropout rate to all layers. See the [Keras
            Dropout layer](https://keras.io/layers/core/#dropout) for more 
            detail.
        - **l2_regs**=`0` : {_tuple_, _list_} of _float_
            - $L_2$-regulatization strength for both the weights and biases
            of the dense layers. A single float will apply the same
            $L_2$-regulatization to all layers.
        """

        # process generic NN hps
        super(DNN, self)._process_hps()

         # required hyperparameters
        self.input_dim = self._proc_arg('input_dim')
        self.dense_sizes = self._proc_arg('dense_sizes')

        # activations
        self.acts = iter_or_rep(self._proc_arg('acts', default='relu'))

        # initializations
        self.k_inits = iter_or_rep(self._proc_arg('k_inits', default='he_uniform'))

        # regularization
        self.dropouts = iter_or_rep(self._proc_arg('dropouts', default=0.))
        self.l2_regs = iter_or_rep(self._proc_arg('l2_regs', default=0.))

        self._verify_empty_hps()

    def _construct_model(self):

        # fresh model
        self._model = Sequential()

        # iterate over specified dense layers
        z = zip(self.dense_sizes, self.acts, self.dropouts, self.l2_regs, self.k_inits)
        looped = False
        for i,(dim, act, dropout, l2_reg, k_init) in enumerate(z):
            looped = True

            # construct variable argument dict
            kwargs = {} if i > 0 else {'input_dim': self.input_dim}
            if l2_reg > 0.:
                kwargs.update({'kernel_regularizer': l2(l2_reg),
                               'bias_regularizer': l2(l2_reg)})

            # add dense layer
            self.model.add(Dense(dim, kernel_initializer=k_init, name=self._proc_name('dense_'+str(i)), **kwargs))
            self._add_act(act)

            # add dropout layer if nonzero
            if dropout > 0.:
                self.model.add(Dropout(dropout, name=self._proc_name('dropout_' + str(i))))

        if not looped:
            raise ValueError('need to specify at least one dense layer')

        # output layer
        self.model.add(Dense(self.output_dim, name=self._proc_name('output')))
        self._add_act(self.output_act)

        # compile model
        self._compile_model()



"""Implementation of EFP/EFM Generator class."""
from __future__ import absolute_import, division, print_function

from collections import Counter
import gzip
import itertools
import json
import time

import numpy as np

from energyflow.algorithms import *
from energyflow.efm import efp2efms
from energyflow.efp import EFP
from energyflow.utils import concat_specs, load_efp_file, transfer
from energyflow.utils.graph_utils import *

igraph = import_igraph()

__all__ = ['Generator']

###############################################################################
# Generator helpers
###############################################################################

def none2inf(x):
    return np.inf if x is None else x

###############################################################################
# Generator
###############################################################################

class Generator(object):

    """Generates non-isomorphic multigraphs according to provided specifications."""

    def __init__(self, dmax=None, nmax=None, emax=None, cmax=None, vmax=None, comp_dmaxs=None,
                      filename=None, gen_efms=True, np_optimize='greedy', verbose=False):
        r"""Doing a fresh generation of connected multigraphs (`filename=None`)
        requires that `igraph` be installed.

        **Arguments**

        - **dmax** : _int_
            - The maximum number of edges of the generated connected graphs.
        - **nmax** : _int_
            - The maximum number of vertices of the generated connected graphs.
        - **emax** : _int_
            - The maximum number of edges of the generated connected simple
            graphs.
        - **cmax** : _int_
            - The maximum VE complexity $\chi$ of the generated connected
            graphs.
        - **vmax** : _int_
            - The maximum valency of the generated connected graphs.
        - **comp_dmaxs** : {_dict_, _int_}
            - If an integer, the maximum number of edges of the generated
            disconnected graphs. If a dictionary, the keys are numbers of
            vertices and the values are the maximum number of edges of the
            generated disconnected graphs with that number of vertices.
        - **filename** : _str_
            - If `None`, do a complete generation from scratch. If set to a
            string, read in connected graphs from the file given, restrict them
            according to the various 'max' parameters, and do a fresh
            disconnected generation. The special value `filename='default'`
            means to read in graphs from the default file. This is useful when
            various disconnected graph parameters are to be varied since the
            generation of large simple graphs is the most computationlly
            intensive part.
        - **gen_efms** : _bool_
            - Controls whether EFM information is generated.
        - **np_optimize** : {`True`, `False`, `'greedy'`, `'optimal'`}
            - The `optimize` keyword of `numpy.einsum_path`.
        - **verbose** : _bool_
            - A flag to control printing.
        """

        start = time.time()

        # check for new generation
        if dmax is not None and filename is None:

            # set maxs
            self._set_maxs(dmax, nmax, emax, cmax, vmax)

            # set options
            self.np_optimize = np_optimize
            self.gen_efms = gen_efms

            # get prime generator instance
            self.pr_gen = PrimeGenerator(self.dmax, self.nmax, self.emax, self.cmax, self.vmax, 
                                         self.gen_efms, self.np_optimize, verbose, start)
            self.cols = self.pr_gen.cols
            self._set_col_inds()

            if verbose:
                print('Finished generating prime graphs in {:.3f}.'.format(time.time() - start))

            # store lists of important quantities
            transfer(self, self.pr_gen, self._prime_attrs())

        # if filename is set, read in file
        else:
            file = load_efp_file(filename)

            # setup cols and col inds
            self.cols = file['cols']
            self._set_col_inds()

            # get maxs from file and passed in options
            c_specs = np.asarray(file['c_specs'])
            for m in ['dmax','nmax','emax','cmax','vmax']:
                setattr(self, m, min(file[m], none2inf(locals()[m])))

            # select connected specs based on maxs
            mask = ((c_specs[:,self.d_ind] <= self.dmax) & 
                    (c_specs[:,self.n_ind] <= self.nmax) & 
                    (c_specs[:,self.e_ind] <= self.emax) & 
                    (c_specs[:,self.c_ind] <= self.cmax) & 
                    (c_specs[:,self.v_ind] <= self.vmax))

            # set ve options
            self.np_optimize = file['np_optimize']

            # get lists of important quantities
            self.gen_efms = file['gen_efms'] and gen_efms
            for attr in (self._prime_attrs()):
                setattr(self, attr, [x for x,m in zip(file[attr],mask) if m])
            self.c_specs = c_specs[mask]


        # setup generator of disconnected graphs
        self._set_comp_dmaxs(comp_dmaxs)
        self.comp_gen = CompositeGenerator(self.c_specs, self.cols, self.comp_dmaxs)

        if verbose:
            print('Finished generating composite graphs in {:.3f}.'.format(time.time() - start))

        # get results and store
        transfer(self, self.comp_gen, self._comp_attrs())

    #################
    # PRIVATE METHODS
    #################

    def _set_col_inds(self):
        self.__dict__.update({col+'_ind': i for i,col in enumerate(self.cols)})

    def _set_maxs(self, dmax, nmax, emax, cmax, vmax):
        self.dmax = dmax
        self.nmax = nmax if nmax is not None else self.dmax + 1
        self.emax = emax if emax is not None else self.dmax
        self.cmax = cmax if cmax is not None else self.nmax
        self.vmax = vmax if vmax is not None else self.dmax

    def _set_comp_dmaxs(self, comp_dmaxs):
        if isinstance(comp_dmaxs, dict):
            self.comp_dmaxs = comp_dmaxs
        else:
            if comp_dmaxs is None:
                comp_dmaxs = self.dmax
            elif not isinstance(comp_dmaxs, int):
                raise TypeError('dmaxs cannot be type {}'.format(type(comp_dmaxs)))

            # implement comp_dmaxs as dict
            self.comp_dmaxs = {}
            if comp_dmaxs >= 2:
                self.comp_dmaxs = {n: comp_dmaxs for n in range(4, 2*comp_dmaxs+1)}

    def _prime_attrs(self, no_global=False):
        attrs = set(['edges', 'weights', 'einstrs', 'einpaths', 'c_specs',
                     'efm_einstrs', 'efm_einpaths', 'efm_specs'])
        return attrs

    def _comp_attrs(self):
        return set(['disc_specs', 'disc_formulae'])

    ################
    # PUBLIC METHODS
    ################

    def save(self, filename, protocol='npz', compression=True):
        """Save the current generator to file.

        **Arguments**

        - **filename** : _str_
            - The path to save the file.
        - **protocol** : {`'npz'`, `'json'`}
            - The file format to be used.
        - **compression** : _bool_
            - Whether to compress the resulting file or not.R
        """
        
        arrs = set(['dmax', 'nmax', 'emax', 'cmax', 'vmax', 'comp_dmaxs',
                    'cols', 'np_optimize', 'gen_efms'])
        arrs |= self._prime_attrs() | self._comp_attrs()
        d = {arr: getattr(self, arr) for arr in arrs}

        if protocol == 'npz':
            if compression:
                np.savez_compressed(filename, **d)
            else:
                np.savez(filename, **d)
        elif protocol == 'json':
            for arr in ['c_specs', 'disc_formulae', 'disc_specs']:
                d[arr] = d[arr].tolist()

            if compression:
                if not filename.endswith('.gz'):
                    filename += '.gz'

                with gzip.open(filename, 'wt') as f:
                    json.dump(d, f)
            else:
                with open(filename, 'wt') as f:
                    json.dump(d, f)
        else:
            raise ValueError('protocol {} not allowed'.format(protocol))

    @property
    def specs(self):
        """An array of EFP specifications. Each row represents an EFP 
        and the columns represent the quantities indicated by `cols`."""
        
        if not hasattr(self, '_specs'):
            self._specs = concat_specs(self.c_specs, self.disc_specs)
        return self._specs

###############################################################################
# PrimeGenerator
###############################################################################

class PrimeGenerator(object):

    """Column descriptions:
    n - number of vertices in graph
    e - number of edges in (underlying) simple graph
    d - number of edges in multigraph
    v - maximum valency of the graph
    k - unique index for graphs with a fixed (n,d)
    c - complexity, with respect to some VE algorithm
    p - number of prime factors for this EFP
    h - number of valency 1 vertices in graph
    """
    cols = ['n','e','d','v','k','c','p','h']

    def __init__(self, dmax, nmax, emax, cmax, vmax, gen_efms, np_optimize, verbose, start):
        """PrimeGenerator __init__."""

        if not igraph:
            raise NotImplementedError('cannot use PrimeGenerator without igraph')
        
        self.ve = VariableElimination(np_optimize)

        # store parameters
        transfer(self, locals(), ['dmax', 'nmax', 'emax', 'cmax', 'vmax', 'gen_efms'])

        # setup N and e values to be used
        self.ns = list(range(1, self.nmax+1))
        self.emaxs = {n: min(self.emax, int(n/2*(n-1))) for n in self.ns}
        self.esbyn = {n: list(range(n-1, self.emaxs[n]+1)) for n in self.ns}

        # this could be more complicated than the same max for all (n,e)
        self.dmaxs = {(n,e): self.dmax for n in self.ns for e in self.esbyn[n]}

        # setup storage containers
        quantities = ['simple_graphs_d', 'edges_d', 'chis_d', 'einpaths_d',
                      'einstrs_d', 'weights_d']
        for q in quantities:
            setattr(self, q, {(n,e): [] for n in self.ns for e in self.esbyn[n]})

        # get simple connected graphs
        self._generate_simple()
        if verbose:
            print('Finished generating simple graphs in {:.3f}.'.format(time.time() - start))

        # get weighted connected graphs
        self._generate_weights()
        if verbose:
            print('Finished generating weighted simple graphs in {:.3f}.'.format(time.time() - start))

        # flatten structures
        self._flatten_structures()
        if verbose:
            print('Finished flattening data structures in {:.3f}.'.format(time.time() - start))

        # efms
        self._generate_efms()
        if verbose:
            print('Finished generating EFMs in {:.3f}.'.format(time.time() - start))

    #################
    # PRIVATE METHODS
    #################

    # generates simple graphs subject to constraints
    def _generate_simple(self):

        self.base_edges = {n: list(itertools.combinations(range(n), 2)) for n in self.ns}

        if self.nmax >= 1:
            self._add_if_new(igraph.Graph.Full(1, directed=False), (1,0))

        # iterate over all combinations of n>1 and d
        for n in self.ns[1:]:
            for e in self.esbyn[n]:

                # consider adding new vertex
                if e-1 in self.esbyn[n-1]:

                    # iterate over all graphs with n-1, e-1
                    for seed_graph in self.simple_graphs_d[(n-1,e-1)]:

                        # iterate over vertices to attach to
                        for v in range(n-1):
                            new_graph = seed_graph.copy()
                            new_graph.add_vertices(1)
                            new_graph.add_edges([(v,n-1)])
                            self._add_if_new(new_graph, (n,e))

                # consider adding new edge to existing set of vertices
                if e-1 in self.esbyn[n]:

                    # iterate over all graphs with n, d-1
                    for seed_graph, seed_edges in zip(self.simple_graphs_d[(n,e-1)], 
                                                      self.edges_d[(n,e-1)]):

                        # iterate over edges that don't exist in graph
                        for new_edge in self._edge_filter(n, seed_edges):
                            new_graph = seed_graph.copy()
                            new_graph.add_edges([new_edge])
                            self._add_if_new(new_graph, (n,e))

    # adds simple graph if it is non-isomorphic to existing graphs and has a valid metric
    def _add_if_new(self, new_graph, ne):

        # check for isomorphism with existing graphs
        for graph in self.simple_graphs_d[ne]:
            if new_graph.isomorphic(graph): 
                return

        # check that ve complexity for this graph is valid
        new_edges = new_graph.get_edgelist()
        einstr, einpath, chi = self.ve.einspecs(new_edges, ne[0])
        if chi > self.cmax: 
            return
        
        # append graph and ve complexity to containers
        self.simple_graphs_d[ne].append(new_graph)
        self.edges_d[ne].append(new_edges)
        self.chis_d[ne].append(chi)

        self.einstrs_d[ne].append(einstr)
        self.einpaths_d[ne].append(einpath)

    # generator for edges not already in list
    def _edge_filter(self, n, edges):
        for edge in self.base_edges[n]:
            if edge not in edges:
                yield edge

    # generates non-isomorphic graph weights subject to constraints
    def _generate_weights(self):

        # take care of the n=2 case
        if (2,1) in self.weights_d:
            self.weights_d[(2,1)].append([(d,) for d in range(1, self.dmaxs[(2,1)]+1)])

        # get ordered integer partitions of d of length e for relevant values
        parts = {}
        for n in self.ns[2:]:
            for e in self.esbyn[n]:
                for d in range(e, self.dmaxs[(n,e)]+1):
                    if (d,e) not in parts:
                        parts[(d,e)] = list(int_partition_ordered(d, e))

        # iterate over the rest of ns
        for n in self.ns[2:]:

            # iterate over es for which there are simple graphs
            for e in self.esbyn[n]:

                # iterate over simple graphs
                for graph in self.simple_graphs_d[(n,e)]:
                    weightings = []

                    # iterate over valid d for this graph
                    for d in range(e, self.dmaxs[(n,e)]+1):

                        # iterate over int partitions
                        for part in parts[(d,e)]:

                            # check that maximum valency is not exceeded 
                            if (self.vmax < self.dmax and 
                                max(graph.strength(weights=part)) > self.vmax):
                                continue

                            # check if isomorphic to existing
                            iso = False
                            for weighting in weightings:
                                if graph.isomorphic_vf2(other=graph, 
                                                        edge_color1=weighting, 
                                                        edge_color2=part): 
                                    iso = True
                                    break
                            if not iso: 
                                weightings.append(part)
                    self.weights_d[(n,e)].append(weightings)

    def _flatten_structures(self):
        c_specs, self.edges, self.weights, self.einstrs, self.einpaths = [], [], [], [], []
        ks = {}

        # handle n=1 case specially
        c_specs.append([1,0,0,0,0,1,1,0])
        self.edges.append(())
        self.weights.append(())
        self.einstrs.append(self.einstrs_d[(1,0)][0])
        self.einpaths.append(self.einpaths_d[(1,0)][0])

        for ne in sorted(self.edges_d.keys()):
            n, e = ne
            z = zip(self.edges_d[ne], self.weights_d[ne], self.chis_d[ne],
                    self.einstrs_d[ne], self.einpaths_d[ne])
            for edgs, ws, c, es, ep in z:
                for w in ws:
                    d = sum(w)
                    k = ks.setdefault((n,d), 0)
                    ks[(n,d)] += 1
                    vs = valencies(EFP(edgs, weights=w).graph).values()
                    v = max(vs)
                    h = Counter(vs)[1]
                    c_specs.append([n, e, d, v, k, c, 1, h])
                    self.edges.append(edgs)
                    self.weights.append(w)
                    self.einstrs.append(es)
                    self.einpaths.append(ep)
        self.c_specs = np.asarray(c_specs)

    def _generate_efms(self):
        self.efm_einstrs, self.efm_specs, self.efm_einpaths = [], [], []
        if self.gen_efms:
            for edgs,ws in zip(self.edges, self.weights):
                einstr, efm_spec = efp2efms(EFP(edgs, weights=ws).graph)
                self.efm_einstrs.append(einstr)
                self.efm_specs.append(efm_spec)
                self.efm_einpaths.append(np.einsum_path(einstr, 
                                                        *[np.empty([4]*sum(s)) for s in efm_spec],
                                                        optimize=self.ve.np_optimize)[0])

###############################################################################
# CompositeGenerator
###############################################################################

class CompositeGenerator(object):

    """CompositeGenerator"""

    def __init__(self, c_specs, cols, comp_dmaxs=None):
        """CompositeGenerator __init__"""

        self.c_specs = c_specs
        self.__dict__.update({col+'_ind': i for i,col in enumerate(cols)})
        self.comp_dmaxs = comp_dmaxs
        
        self.ns = sorted(self.comp_dmaxs.keys())
        self.nmax_avail = np.max(self.c_specs[:,self.n_ind]) if len(self.c_specs) else 0

        self.ks, self.ndk2i = {}, {}
        for i,spec in enumerate(self.c_specs):
            n, d, k = spec[[self.n_ind, self.d_ind, self.k_ind]]
            self.ks.setdefault((n,d), 0)
            self.ks[(n,d)] += 1
            self.ndk2i[(n,d,k)] = i

        self._generate_disconnected()

    #################
    # PRIVATE METHODS
    #################

    def _generate_disconnected(self):
        
        disc_formulae, disc_specs = [], []

        for n in self.ns:

            # partitions with no 1s, no numbers > self.nmax_avail, and not the trivial partition
            good_part = lambda x: (1 not in x and max(x) <= self.nmax_avail and len(x) > 1)
            n_parts = [tuple(x) for x in int_partition_unordered(n) if good_part(x)]
            n_parts.sort(key=len)

            # iterate over all ds
            for d in range(int((n-1)/2)+1, self.comp_dmaxs[n]+1):

                # iterate over all n_parts
                for n_part in n_parts:
                    n_part_len = len(n_part)

                    # get d_parts of the right length
                    d_parts = [x for x in int_partition_unordered(d) if len(x) == n_part_len]

                    # ensure that we found some
                    if len(d_parts) == 0: continue

                    # usage of set and sorting is important to avoid duplicates
                    specs = set()

                    # iterate over all orderings of the n_part
                    for n_part_ord in set(itertools.permutations(n_part)):

                        # iterate over all d_parts
                        for d_part in d_parts:

                            # construct spec. sorting ensures we don't get duplicates in specs
                            spec = tuple(sorted([(npo,dp) for npo,dp in zip(n_part_ord,d_part)]))

                            # check that we have the proper primes to calculate this spec
                            good = True
                            for pair in spec:
                                if pair not in self.ks:
                                    good = False
                                    break
                            if good:
                                specs.add(spec)

                    # iterate over all specs that we found
                    for spec in specs:

                        # keep track of how many we added
                        kcount = 0 if (n,d) not in self.ks else self.ks[(n,d)]

                        # iterate over all possible formula implementations with the different ndk
                        for kspec in itertools.product(*[range(self.ks[factor]) for factor in spec]):

                            # iterate over factors
                            formula = []
                            cmax = e = vmax = h = 0 
                            for (nn,dd),kk in zip(spec,kspec):

                                # add (n,d,k) of factor to formula
                                ndk = (nn,dd,kk)
                                formula.append(ndk)

                                # select original simple graph
                                ind = self.ndk2i[ndk]
                                cmax = max(cmax, self.c_specs[ind, self.c_ind])
                                e += self.c_specs[ind, self.e_ind]
                                vmax = max(vmax, self.c_specs[ind, self.v_ind])
                                h += self.c_specs[ind, self.h_ind]

                            # append to stored array
                            disc_formulae.append(tuple(sorted(formula)))
                            disc_specs.append([n, e, d, vmax, kcount, cmax, len(kspec), h])
                            kcount += 1

        # ensure unique formulae (deals with possible degeneracy in selection of factors)
        disc_form_set = set()
        mask = [not(form in disc_form_set or disc_form_set.add(form)) for form in disc_formulae]

        # store as numpy arrays
        self.disc_formulae = np.asarray(disc_formulae)[mask]
        self.disc_specs = np.asarray(disc_specs)[mask]


from __future__ import absolute_import, division, print_function

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.linear_model import LogisticRegression

from energyflow.archs.archbase import ArchBase

__all__ = ['LinearClassifier']

###############################################################################
# LinearClassifier
###############################################################################

class LinearClassifier(ArchBase):

    """Linear classifier that can be either Fisher's linear discriminant
    or logistic regression. Relies on the [scikit-learn](https://scikit-learn.org/)
    implementations of these classifiers."""

    # LinearClassifier(*args, **kwargs)
    def _process_hps(self):
        """See [`ArchBase`](#archbase) for how to pass in hyperparameters.

        **Default Hyperparameters**

        - **linclass_type**=`'lda'` : {`'lda'`, `'lr'`}
            - Controls which type of linear classifier is used. `'lda'`
            corresponds to [`LinearDisciminantAnalysis`](http://scikit-
            learn.org/stable/modules/generated/sklearn.discriminant_analysis.
            LinearDiscriminantAnalysis.html) and `'lr'` to [`Logistic
            Regression`](http://scikit-learn.org/stable/modules/generated/
            sklearn.linear_model.LogisticRegression.html). If using `'lr'`
            all arguments are passed on directly to the scikit-learn
            class.

        **Linear Discriminant Analysis Hyperparameters**

        - **solver**=`'svd'` : {`'svd'`, `'lsqr'`, `'eigen'`}
            - Which LDA solver to use.
        - **tol**=`1e-12` : _float_
            - Threshold used for rank estimation. Notably not a
            convergence parameter.

        **Logistic Regression Hyperparameters**

        - **LR_hps**=`{}` : _dict_
            - Dictionary of keyword arguments to pass on to the underlying
            `LogisticRegression` model.
        """

        # which type of linear model we're using
        self.linclass_type = self.hps.get('linclass_type', 'lda')

        # LDA hyperparameters
        self.solver = self.hps.get('solver', 'svd')
        self.tol = self.hps.get('tol', 10**-12)

        # logistic regression hyperparameter dictionary
        self.LR_hps = self.hps.get('LR_hps', {})

    def _construct_model(self):

        # setup linear model according to linclass_type
        if self.linclass_type == 'lda':
            self._model = LinearDiscriminantAnalysis(solver=self.solver, tol=self.tol)
        elif self.linclass_type == 'lr':
            self._model = LogisticRegression(**self.LR_hps)
        else:
            raise ValueError('linclass_type can only be lda or lr')

    def fit(self, *args, **kwargs):
        return self.model.fit(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self.model.predict_proba(*args, **kwargs)

    @property
    def model(self):
        return self._model



"""Utility functions for various purposes relating to the uses of EnergyFlow."""
from __future__ import absolute_import

from . import data_utils
from . import event_utils
from . import fastjet_utils
from . import generic_utils
from . import graph_utils
from . import image_utils
from . import particle_utils

from .data_utils import *
from .event_utils import *
from .fastjet_utils import *
from .generic_utils import *
from .graph_utils import *
from .image_utils import *
from .particle_utils import *

__all__ = (event_utils.__all__ + 
           fastjet_utils.__all__ +
           particle_utils.__all__)



r"""Energy Flow Networks (EFNs) and Particle Flow Networks (PFNs) are model
architectures designed for learning from collider events as unordered,
variable-length sets of particles. Both EFNs and PFNs are parameterized by a
learnable per-particle function $\Phi$ and latent space function $F$.

An EFN takes the following form:
$$\text{EFN}=F\left(\sum_{i=1}^M z_i \Phi(\hat p_i)\right)$$
where $z_i$ is a measure of the energy of particle $i$, such as $z_i=p_{T,i}$,
and $\hat p_i$ is a measure of the angular information of particle $i$, such as
$\hat p_i = (y_i,\phi_i)$. Any infrared- and collinear-safe observable can be
parameterized in this form.

A PFN takes the following form:
$$\text{PFN}=F\left(\sum_{i=1}^M \Phi(p_i)\right)$$
where $p_i$ is the information of particle $i$, such as its four-momentum,
charge, or flavor. Any observable can be parameterized in this form. See the
[Deep Sets](https://arxiv.org/abs/1703.06114) framework for additional 
discussion.

Since these architectures are not used by the core EnergyFlow code, and require
the external [Keras](https://keras.io) and [scikit-learn](http://scikit-
learn.org/) libraries, they are not imported by default but must be explicitly
imported, e.g. `from energyflow.archs import *`. EnergyFlow also contains
several additional model architectures for ease of using common models that
frequently appear in the intersection of particle physics and machine learning.
"""
from __future__ import absolute_import

import warnings

__all__ = []

# requires keras
try:
    from . import cnn
    from . import dnn
    from . import efn
    from .cnn import *
    from .dnn import *
    from .efn import *

    __all__ += cnn.__all__ + dnn.__all__ + efn.__all__

except ImportError as e:
    warnings.warn('could not import some architectures - ' + str(e))

# requires sklearn
try:
    from . import linear
    from .linear import *

    __all__ += linear.__all__

except ImportError as e:
    warnings.warn('could not import some architectures - ' + str(e))



r"""### Particle Tools

Tools for dealing with particle momenta four-vectors. A four-vector can either
be in Cartesian coordinates, `[e,px,py,pz]` (energy, momentum in `x` direction,
momentum in `y` direction, momentum in `z` direction), or hadronic coordinates, 
`[pt,y,phi,m]` (transverse momentum, rapidity, azimuthal angle, mass), which
are related via:

$$p_T=\sqrt{p_x^2+p_y^2},\quad y=\text{arctanh}\,\frac{p_z}{E},\quad 
\phi=\arctan_2\frac{p_y}{p_x},\quad m=\sqrt{E^2-p_x^2-p_y^2-p_z^2}$$

and inversely:

$$E=\cosh y\sqrt{p_T^2+m^2},\quad p_x=p_T\cos\phi,\quad 
p_y=p_T\sin\phi,\quad p_z=\sinh y\sqrt{p_T^2+m^2}.$$

The pseudorapidity `eta` can be obtained from a Cartesian four-momentum as:

$$\eta=\text{arctanh}\,\frac{p_z}{|\vec p|},\quad 
|\vec p|\equiv\sqrt{p_x^2+p_y^2+p_z^2},$$

and is related to the rapidity via

$$\eta=\text{arcsinh}\left(\sinh y\,\left(1+m^2/p_T^2\right)^{1/2}\right),\quad 
y=\text{arcsinh}\left(\sinh \eta\,\left(1+m^2/p_T^2\right)^{-1/2}\right).$$

Note that the above formulas are numerically stable up to values of rapidity or
pseudorapidity of a few hundred, above which the formulas have numerical issues. 
In this case, a different but equivalent formulae are used that are numerically
stable in this region. In all cases, the $p_T\to0$ limit produces infinite
values.

In the context of this package, an "event" is a two-dimensional numpy array
with shape `(M,4)` where `M` is the multiplicity. An array of events is a 
three-dimensional array with shape `(N,M,4)` where `N` is the number of events.
The valid inputs and outputs of the functions here will be described using
this terminology.
"""
from __future__ import absolute_import, division, print_function

import warnings

import numpy as np
import six

__all__ = [

    # from_p4s functions
    'ptyphims_from_p4s',
    'pts_from_p4s',
    'pt2s_from_p4s',
    'ys_from_p4s',
    'etas_from_p4s',
    'phis_from_p4s',
    'm2s_from_p4s',
    'ms_from_p4s',
    'ms_from_ps',

    # eta/y conversions
    'etas_from_pts_ys_ms',
    'ys_from_pts_etas_ms',

    # from_ptyphims functions
    'p4s_from_ptyphims',
    'p4s_from_ptyphipids',

    # combination functions
    'sum_ptyphims',
    'sum_ptyphipids',

    # transformation functions
    'center_ptyphims',
    'rotate_ptyphims',
    'reflect_ptyphims',

    # pid functions
    'pids2ms',
    'pids2chrgs',
    'ischrgd',

    # other functions
    'phi_fix',
    'flat_metric',
]

def ptyphims_from_p4s(p4s, phi_ref=None):
    r"""Convert to hadronic coordinates `[pt,y,phi,m]` from Cartesian
    coordinates. All-zero four-vectors are left alone.

    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian coordinates.
    - **phi_ref** : {`None`, `'hardest'`, _float_, _numpy.ndarray_}
        - Used to help deal with the fact that $\phi$ is a periodic coordinate.
        If a float (which should be in $[0,2\pi)$), all phi values will be
        within $\pm\pi$ of this reference value. If `'\hardest'`, the phi of
        the hardest particle is used as the reference value. If `None`, all
        phis will be in the range $[0,2\pi)$. An array is accepted in the case
        that `p4s` is an array of events, in which case the `phi_ref` array
        should have shape `(N,)` where `N` is the number of events.

    **Returns**

    - _numpy.ndarray_
        - An array of hadronic four-momenta with the same shape as the input.
    """

    p4s = np.asarray(p4s, dtype=float)
    if p4s.shape[-1] != 4:
        raise ValueError("Last dimension of 'p4s' must have size 4.")

    out = np.zeros(p4s.shape, dtype=float)
    out[...,0] = pts_from_p4s(p4s)
    out[...,1] = ys_from_p4s(p4s)
    out[...,2] = phis_from_p4s(p4s, phi_ref, _pts=out[...,0])
    out[...,3] = ms_from_p4s(p4s)

    return out

def pt2s_from_p4s(p4s):
    """Calculate the squared transverse momenta of a collection of four-vectors.
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian coordinates.

    **Returns**

    - _numpy.ndarray_
        - An array of squared transverse momenta with shape `p4s.shape[:-1]`.
    """

    p4s = np.asarray(p4s, dtype=float)
    return p4s[...,1]**2 + p4s[...,2]**2

def pts_from_p4s(p4s):
    """Calculate the transverse momenta of a collection of four-vectors.
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian coordinates.

    **Returns**

    - _numpy.ndarray_
        - An array of transverse momenta with shape `p4s.shape[:-1]`.
    """

    return np.sqrt(pt2s_from_p4s(p4s))

def ys_from_p4s(p4s):
    """Calculate the rapidities of a collection of four-vectors. Returns zero
    for all-zero particles
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian coordinates.

    **Returns**

    - _numpy.ndarray_
        - An array of rapidities with shape `p4s.shape[:-1]`.
    """

    p4s = np.asarray(p4s, dtype=float)
    out = np.zeros(p4s.shape[:-1], dtype=float)

    nz_mask = np.any(p4s != 0., axis=-1)
    nz_p4s = p4s[nz_mask]
    out[nz_mask] = np.arctanh(nz_p4s[...,3]/nz_p4s[...,0])

    return out

def etas_from_p4s(p4s):
    """Calculate the pseudorapidities of a collection of four-vectors. Returns
    zero for all-zero particles
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian coordinates.

    **Returns**

    - _numpy.ndarray_
        - An array of pseudorapidities with shape `p4s.shape[:-1]`.
    """

    p4s = np.asarray(p4s, dtype=float)
    out = np.zeros(p4s.shape[:-1], dtype=float)

    nz_mask = np.any(p4s != 0., axis=-1)
    nz_p4s = p4s[nz_mask]
    out[nz_mask] = np.arctanh(nz_p4s[...,3]/np.sqrt(nz_p4s[...,1]**2 + nz_p4s[...,2]**2 + nz_p4s[...,3]**2))

    return out

# phis_from_p4s(p4s, phi_ref=None)
def phis_from_p4s(p4s, phi_ref=None, _pts=None):
    r"""Calculate the azimuthal angles of a collection of four-vectors.
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian coordinates.
    - **phi_ref** : {_float_, _numpy.ndarray_, `None`, `'hardest'`}
        - Used to help deal with the fact that $\phi$ is a periodic coordinate.
        If a float (which should be in $[0,2\pi)$), all phi values will be
        within $\pm\pi$ of this reference value. If `'\hardest'`, the phi of
        the hardest particle is used as the reference value. If `None`, all
        phis will be in the range $[0,2\pi)$. An array is accepted in the case
        that `p4s` is an array of events, in which case the `phi_ref` array
        should have shape `(N,)` where `N` is the number of events.

    **Returns**

    - _numpy.ndarray_
        - An array of azimuthal angles with shape `p4s.shape[:-1]`.
    """

    # get phis
    p4s = np.asarray(p4s, dtype=float)
    phis = np.asarray(np.arctan2(p4s[...,2], p4s[...,1]))
    phis[phis<0] += 2*np.pi

    # ensure close to reference value
    if phi_ref is not None:
        if isinstance(phi_ref, six.string_types) and phi_ref == 'hardest':
            ndim = phis.ndim

            # here the particle is already phi fixed with respect to itself
            if ndim == 0:
                return phis

            # get pts if needed (pt2s are fine for determining hardest)
            if _pts is None:
                _pts = pt2s_from_p4s(p4s)
            hardest = np.argmax(_pts, axis=-1)

            # indexing into vector
            if ndim == 1:
                phi_ref = phis[hardest]

            # advanced indexing
            elif ndim == 2:
                phi_ref = phis[np.arange(len(phis)), hardest]

            else:
                raise ValueError("'p4s' should not have more than three dimensions.")

        phis = phi_fix(phis, phi_ref, copy=False)

    return phis

TWOPI = 2*np.pi
def phi_fix(phis, phi_ref, copy=True):
    r"""A function to ensure that all phis are within $\pi$ of `phi_ref`. It is
    assumed that all starting phi values are $\pm 2\pi$ of `phi_ref`.

    **Arguments**

    - **phis** : _numpy.ndarray_ or _list_
        - Array of phi values.
    - **phi_ref** : {_float_ or _numpy.ndarray_}
        - A reference value used so that all phis will be within $\pm\pi$ of
        this value. Should have a shape of `phis.shape[:-1]`.
    - **copy** : _bool_
        - Determines if `phis` are copied or not. If `False` then `phis` is
        modified in place.

    **Returns**

    - _numpy.ndarray_
        - An array of the fixed phi values.
    """

    phis, phi_ref = np.asarray(phis, dtype=float), np.asarray(phi_ref, dtype=float)
    phi_ref = phi_ref[...,np.newaxis] if phi_ref.ndim > 0 else phi_ref

    diff = phis - phi_ref

    new_phis = np.copy(phis) if copy else phis
    new_phis[diff > np.pi] -= TWOPI
    new_phis[diff < -np.pi] += TWOPI

    return new_phis

def m2s_from_p4s(p4s):
    """Calculate the squared masses of a collection of four-vectors.
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian coordinates.

    **Returns**

    - _numpy.ndarray_
        - An array of squared masses with shape `p4s.shape[:-1]`.
    """

    p4s = np.asarray(p4s, dtype=float)
    return p4s[...,0]**2 - p4s[...,1]**2 - p4s[...,2]**2 - p4s[...,3]**2

def ms_from_p4s(p4s):
    """Calculate the masses of a collection of four-vectors.
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian coordinates.

    **Returns**

    - _numpy.ndarray_
        - An array of masses with shape `p4s.shape[:-1]`.
    """

    m2s = m2s_from_p4s(p4s)
    return np.sign(m2s)*np.sqrt(np.abs(m2s))

def ms_from_ps(ps):
    r"""Calculate the masses of a collection of Lorentz vectors in two or more
    spacetime dimensions.

    **Arguments**

    - **ps** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in Cartesian
        coordinates in $d\ge2$ spacetime dimensions.

    **Returns**

    - _numpy.ndarray_
        - An array of masses with shape `ps.shape[:-1]`.
    """

    nps = np.asarray(ps, dtype=float)
    m2s = nps[...,0]**2 - np.sum(nps[...,1:]**2, axis=-1)
    return np.sign(m2s)*np.sqrt(np.abs(m2s))

# etas_from_pts_ys_ms(pts, ys, ms)
def etas_from_pts_ys_ms(pts, ys, ms, _cutoff=50.):
    """Calculate pseudorapidities from transverse momenta, rapidities, and masses.
    All input arrays should have the same shape.

    **Arguments**

    - **pts** : _numpy.ndarray_
        - Array of transverse momenta.
    - **ys** : _numpy.ndarray_
        - Array of rapidities.
    - **ms** : _numpy.ndarray_
        - Array of masses.

    **Returns**

    - _numpy.ndarray_
        - Array of pseudorapidities with the same shape as `ys`.
    """

    pts, ys, ms = np.asarray(pts), np.asarray(ys), np.asarray(ms)

    abs_ys, x2s = np.abs(ys), (ms/pts)**2
    sqrt1plusx2s = np.sqrt(1 + x2s)

    if np.max(abs_ys) < _cutoff:
        return np.arcsinh(np.sinh(ys)*sqrt1plusx2s)

    # have to use different formulas for large and small ys
    large_mask = (abs_ys > _cutoff)
    small_mask = ~large_mask
    out = np.zeros(ys.shape, dtype=float)

    large_abs_ys = abs_ys[large_mask]

    # note that the commented term can be ignored since it is numerically 1 for |y| > 20
    out[large_mask] = large_abs_ys + np.log(#(1. - np.exp(-2.*large_abs_ys))*
                        (sqrt1plusx2s[large_mask] + 
                         np.sqrt(x2s[large_mask] + 1./np.tanh(large_abs_ys)**2))/2.)
    out[large_mask] *= np.sign(ys[large_mask])

    out[small_mask] = np.arcsinh(np.sinh(ys[small_mask])*sqrt1plusx2s[small_mask])

    return out

# ys_from_pts_etas_ms(pts, etas, ms)
def ys_from_pts_etas_ms(pts, etas, ms, _cutoff=50.):
    """Calculate rapidities from transverse momenta, pseudorapidities, and masses.
    All input arrays should have the same shape.

    **Arguments**

    - **pts** : _numpy.ndarray_
        - Array of transverse momenta.
    - **etas** : _numpy.ndarray_
        - Array of pseudorapidities.
    - **ms** : _numpy.ndarray_
        - Array of masses.

    **Returns**

    - _numpy.ndarray_
        - Array of rapidities with the same shape as `etas`.
    """

    pts, etas, ms = np.asarray(pts), np.asarray(etas), np.asarray(ms)

    abs_etas, x2s = np.abs(etas), (ms/pts)**2
    sqrt1plusx2s = np.sqrt(1 + x2s)

    if np.max(abs_etas) < _cutoff:
        return np.arcsinh(np.sinh(etas)/sqrt1plusx2s)

    # have to use different formulas for large and small etas
    large_mask = (abs_etas > _cutoff)
    small_mask = ~large_mask
    out = np.zeros(etas.shape, dtype=float)

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        large_abs_etas = abs_etas[large_mask]

        # note that the commented term can be ignored since it is numerically 1 for |eta| > 20
        out[large_mask] = large_abs_etas + np.log(#(1. - np.exp(-2.*large_abs_etas))*
                                   (1. + np.sqrt(1./np.tanh(large_abs_etas)**2 + 
                                                 x2s[large_mask]/np.sinh(large_abs_etas)**2))/
                                   (2.*sqrt1plusx2s[large_mask]))
        out[large_mask] *= np.sign(etas[large_mask])

    out[small_mask] = np.arcsinh(np.sinh(etas[small_mask])/sqrt1plusx2s[small_mask])

    return out

def p4s_from_ptyphims(ptyphims):
    """Calculate Cartesian four-vectors from transverse momenta, rapidities,
    azimuthal angles, and (optionally) masses for each input.
    
    **Arguments**

    - **ptyphims** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in hadronic coordinates.
        The mass is optional and if left out will be taken to be zero.

    **Returns**

    - _numpy.ndarray_
        - An array of Cartesian four-vectors.
    """

    # get pts, ys, phis
    ptyphims = np.asarray(ptyphims, dtype=float)
    pts, ys, phis = (ptyphims[...,0,np.newaxis], 
                     ptyphims[...,1,np.newaxis], 
                     ptyphims[...,2,np.newaxis])

    # get masses
    ms = ptyphims[...,3,np.newaxis] if ptyphims.shape[-1] == 4 else np.zeros(pts.shape)

    Ets = np.sqrt(pts**2 + ms**2)
    p4s = np.concatenate((Ets*np.cosh(ys), pts*np.cos(phis), 
                          pts*np.sin(phis), Ets*np.sinh(ys)), axis=-1)
    return p4s

def p4s_from_ptyphipids(ptyphipids, error_on_unknown=False):
    """Calculate Cartesian four-vectors from transverse momenta, rapidities,
    azimuthal angles, and particle IDs for each input. The particle IDs are
    used to lookup the mass of the particle. Transverse momenta should have
    units of GeV when using this function.
    
    **Arguments**

    - **ptyphipids** : _numpy.ndarray_ or _list_
        - A single particle, event, or array of events in hadronic coordinates
        where the mass is replaced by the PDG ID of the particle.
    - **error_on_unknown** : _bool_
        - See the corresponding argument of [`pids2ms`](#pids2ms).

    **Returns**

    - _numpy.ndarray_
        - An array of Cartesian four-vectors with the same shape as the input.
    """

    # get pts, ys, phis
    ptyphipids = np.asarray(ptyphipids, dtype=float)
    pts, ys, phis = (ptyphipids[...,0,np.newaxis],
                     ptyphipids[...,1,np.newaxis],
                     ptyphipids[...,2,np.newaxis])

    # get masses
    ms = pids2ms(ptyphipids[...,3,np.newaxis], error_on_unknown)

    Ets = np.sqrt(pts**2 + ms**2)
    p4s = np.concatenate((Ets*np.cosh(ys), pts*np.cos(phis), 
                          pts*np.sin(phis), Ets*np.sinh(ys)), axis=-1)
    return p4s

def sum_ptyphims(ptyphims, scheme='escheme'):
    r"""Add a collection of four-vectors that are expressed in hadronic
    coordinates by first converting to Cartesian coordinates and then summing.

    **Arguments**

    - **ptyphims** : _numpy.ndarray_ or _list_
        - An event in hadronic coordinates. The mass is optional and if left
        out will be taken to be zero.
    - **scheme** : _str_
        - A string specifying a recombination scheme for adding four-vectors
        together. Currently supported options are `'escheme'`, which adds the
        vectors in Cartesian coordinates, and `'ptscheme'`, which sums the pTs
        of each particle and places the jet axis at the pT-weighted centroid
        in the rapidity-azimuth plane. Note that `'ptscheme'` will return a
        three-vector consisting of the jet `[pT,y,phi]` with no mass value.

    **Returns**

    - _numpy.ndarray_
        - Array of summed four-vectors, in hadronic coordinates. Note that when
        `scheme` is `'escheme'`, the $\phi$ value of the hardest particle is
        used as the `phi_ref` when converting back to hadronic coordinates.
    """

    ptyphims = np.asarray(ptyphims, dtype=float)

    if ptyphims.ndim <= 1 or ptyphims.size == 0:
        return ptyphims

    if scheme == 'escheme':
        phi = ptyphims[np.argmax(ptyphims[:,0]),2]
        sum_p4 = np.sum(p4s_from_ptyphims(ptyphims), axis=-2)
        return ptyphims_from_p4s(sum_p4, phi_ref=phi)

    elif scheme == 'ptscheme':
        y, phi = np.average(ptyphims[:,1:3], weights=ptyphims[:,0], axis=0)
        return np.asarray([np.sum(ptyphims[:,0]), y, phi])

    else:
        raise ValueError('Unknown recombination scheme {}'.format(scheme))

def sum_ptyphipids(ptyphipids, scheme='escheme', error_on_unknown=False):
    r"""Add a collection of four-vectors that are expressed as
    `[pT,y,phi,pdgid]`.

    **Arguments**

    - **ptyphipids** : _numpy.ndarray_ or _list_
        - A single particle or event in hadronic coordinates where the mass
        is replaced by the PDG ID of the particle.
    - **scheme** : _str_
        - See the argument of the same name of [`sum_ptyphims`](#sum_ptyphims).
    - **error_on_unknown** : _bool_
        - See the corresponding argument of [`pids2ms`](#pids2ms).

    **Returns**

    - _numpy.ndarray_
        - Array of summed four-vectors, in hadronic coordinates. Note that when
        `scheme` is `'escheme'`, the $\phi$ value of the hardest particle is
        used as the `phi_ref` when converting back to hadronic coordinates.
    """

    ptyphipids = np.asarray(ptyphipids, dtype=float)

    if ptyphipids.ndim <= 1 or ptyphipids.size == 0:
        return ptyphipids

    if scheme == 'escheme':
        phi = ptyphipids[np.argmax(ptyphipids[:,0]),2]
        sum_p4 = np.sum(p4s_from_ptyphipids(ptyphipids, error_on_unknown), axis=-2)
        return ptyphims_from_p4s(sum_p4, phi_ref=phi)

    elif scheme == 'ptscheme':
        return sum_ptyphims(ptyphipids, scheme=scheme)

    else:
        raise ValueError('Unknown recombination scheme {}'.format(scheme))

def center_ptyphims(ptyphims, axis=None, center='escheme', copy=True):
    """Center a collection of four-vectors according to a calculated or 
    provided axis.

    **Arguments**

    - **ptyphims** : _numpy.ndarray_ or _list_
        - An event in hadronic coordinates. The mass is optional and if left
        out will be taken to be zero.
    - **axis** : _numpy.ndarray_
        - If not `None`, the `[y,phi]` values to use for centering.
    - **center** : _str_
        - The centering scheme to be used. Valid options are the same as the
        `scheme` argument of [`sum_ptyphims`](#sum_ptyphims).
    - **copy** : _bool_
        - Whether or not to copy the input array.

    **Returns**

    - _numpy.ndarray_
        - An array of hadronic four-momenta with the positions centered around
        the origin.
    """

    if axis is None:
        axis = sum_ptyphims(ptyphims, scheme=center)[1:3]

    if copy:
        ptyphims = np.copy(ptyphims)

    ptyphims[:,1:3] -= axis

    return ptyphims

def _do_reflection(zs, coords):
    return np.sum(zs[coords > 0.]) < np.sum(zs[coords < 0.])

def rotate_ptyphims(ptyphims, rotate='ptscheme', center=None, copy=True):
    """Rotate a collection of four-vectors to vertically align the principal
    component of the energy flow. The principal component is obtained as the
    eigenvector of the energy flow with the largest eigenvalue. It is only
    defined up to a sign, however it is ensured that there is more total pT in 
    the top half of the rapidity-azimuth plane.

    **Arguments**

    - **ptyphims** : _numpy.ndarray_ or _list_
        - An event in hadronic coordinates. The mass is optional and if left
        out will be taken to be zero.
    - **rotate** : _str_
        - The rotation scheme to be used. Currently, only `'ptscheme'` is
        supported, which causes the rotation to take place in the 
        rapidity-azimuth plane.
    - **center** : _str_ or `None`
        - If not `None`, the event will be centered prior to rotation and this
        argument will be passed on to `center_ptyphims` as the centering
        scheme.
    - **copy** : _bool_
        - Whether or not to copy the input array.

    **Returns**

    - _numpy.ndarray_
        - An array of hadronic four-momenta with the positions rotated around
        the origin.
    """

    if copy:
        ptyphims = np.copy(ptyphims)

    if center is not None:
        ptyphims = center_ptyphims(ptyphims, center=center, copy=False)

    if rotate == 'ptscheme':

        zs, phats = ptyphims[:,0], ptyphims[:,1:3]
        efm2 = np.einsum('a,ab,ac->bc', zs, phats, phats, optimize=['einsum_path', (0,1), (0,1)])
        eigvals, eigvecs = np.linalg.eigh(efm2)

        ptyphims[:,1:3] = np.dot(phats, eigvecs)

        if _do_reflection(zs, ptyphims[:,2]):
            ptyphims[:,1:3] *= -1.

    else:
        raise ValueError('Unknown rotation scheme {}'.format(rotate))

    return ptyphims

def reflect_ptyphims(ptyphims, which='both', center=None, copy=True):
    """Reflect a collection of four-vectors to arrange the highest-pT
    half or quadrant to have positive rapidity-azimuth coordinates.

    **Arguments**

    - **ptyphims** : _numpy.ndarray_
        - An event in hadronic coordinates. The mass is optional and is not
        used as a part of this function.
    - **which** : {`'both'`, `'x'`, `'y'`}
        - Controls which axes to consider reflecting over. `'both'` includes
        `'x'` and `'y'`.
    - **center** : _str_ or `None`
        - If not `None`, the centering scheme to use prior to performing
        reflections.
    - **copy** : _bool_
        - Whether or not to copy the input array.
    """

    if copy:
        ptyphims = np.copy(ptyphims)

    if center is not None:
        ptyphims = center_ptyphims(ptyphims, center=center, copy=False)

    zs = ptyphims[:,0]
    if (which == 'both' or which == 'x') and _do_reflection(zs, ptyphims[:,1]):
        ptyphims[:,1] *= -1.

    if (which == 'both' or which == 'y') and _do_reflection(zs, ptyphims[:,2]):
        ptyphims[:,2] *= -1.

    return ptyphims

# masses (in GeV) of particles by pdgid
# obtained from the Pythia8 Particle Data page
# http://home.thep.lu.se/~torbjorn/pythia82html/ParticleData.html
# includes fundamental particles and most ground state uds mesons and baryons
PARTICLE_MASSES = {
    0:    0.,      # void
    1:    0.33,    # down
    2:    0.33,    # up
    3:    0.50,    # strange
    4:    1.50,    # charm
    5:    4.80,    # bottom
    6:    171.,    # top
    11:   5.11e-4, # e-
    12:   0.,      # nu_e
    13:   0.10566, # mu-
    14:   0.,      # nu_mu
    15:   1.77682, # tau-
    16:   0.,      # nu_tau
    21:   0.,      # gluon
    22:   0.,      # photon
    23:   91.1876, # Z
    24:   80.385,  # W+
    25:   125.,    # Higgs
    111:  0.13498, # pi0
    113:  0.77549, # rho0
    130:  0.49761, # K0-long
    211:  0.13957, # pi+
    213:  0.77549, # rho+
    221:  0.54785, # eta
    223:  0.78265, # omega
    310:  0.49761, # K0-short
    321:  0.49368, # K+
    331:  0.95778, # eta'
    333:  1.01946, # phi
    2112: 0.93957, # neutron
    2212: 0.93827, # proton
    1114: 1.232,   # Delta-
    2114: 1.232,   # Delta0
    2214: 1.232,   # Delta+
    2224: 1.232,   # Delta++
    3122: 1.11568, # Lambda0
    3222: 1.18937, # Sigma+
    3212: 1.19264, # Sigma0
    3112: 1.19745, # Sigma-
    3312: 1.32171, # Xi-
    3322: 1.31486, # Xi0
    3334: 1.67245, # Omega-
}

# particle electric charges in fundamental units
PARTICLE_CHARGES = {
    0:     0.,   # void
    1:    -1./3, # down
    2:     2./3, # up
    3:    -1./3, # strange
    4:     2./3, # charm
    5:    -1./3, # bottom
    6:     2./3, # top
    11:   -1.,   # e-
    12:    0.,   # nu_e
    13:   -1.,   # mu-
    14:    0.,   # nu_mu
    15:   -1.,   # tau-
    16:    0.,   # nu_tau
    21:    0.,   # gluon
    22:    0.,   # photon
    23:    0.,   # Z
    24:    1.,   # W+
    25:    0.,   # Higgs
    111:   0.,   # pi0
    113:   0.,   # rho0
    130:   0.,   # K0-long
    211:   1.,   # pi+
    213:   1.,   # rho+
    221:   0.,   # eta
    223:   0.,   # omega
    310:   0.,   # K0-short
    321:   1.,   # K+
    331:   0.,   # eta'
    333:   0.,   # phi
    2112:  0.,   # neutron
    2212:  1.,   # proton
    1114: -1.,   # Delta-
    2114:  0.,   # Delta0
    2214:  1.,   # Delta+
    2224:  2.,   # Delta++
    3122:  0.,   # Lambda0
    3222:  1.,   # Sigma+
    3212:  0.,   # Sigma0
    3112: -1.,   # Sigma-
    3312: -1.,   # Xi-
    3322:  0.,   # Xi0
    3334: -1.,   # Omega-
}

CHARGED_PIDS = frozenset(k for k,v in PARTICLE_CHARGES.items() if v != 0.)

def pids2ms(pids, error_on_unknown=False):
    r"""Map an array of [Particle Data Group IDs](http://pdg.lbl.gov/2018/
    reviews/rpp2018-rev-monte-carlo-numbering.pdf) to an array of the
    corresponding particle masses (in GeV).

    **Arguments**

    - **pids** : _numpy.ndarray_ or _list_
        - An array of numeric (float or integer) PDG ID values.
    - **error_on_unknown** : _bool_
        - Controls whether a `KeyError` is raised if an unknown PDG ID is
        encountered. If `False`, unknown PDG IDs will map to zero.

    **Returns**

    - _numpy.ndarray_
        - An array of masses in GeV.
    """

    abspids = np.abs(np.asarray(pids, dtype=int))
    orig_shape = abspids.shape
    abspids = abspids.reshape(-1)

    if error_on_unknown:
        masses = [PARTICLE_MASSES[pid] for pid in abspids]
    else:
        masses = [PARTICLE_MASSES.get(pid, 0.) for pid in abspids]

    return np.asarray(masses, dtype=float).reshape(orig_shape)

def pids2chrgs(pids, error_on_unknown=False):
    r"""Map an array of [Particle Data Group IDs](http://pdg.lbl.gov/2018/
    reviews/rpp2018-rev-monte-carlo-numbering.pdf) to an array of the
    corresponding particle charges (in fundamental units where the charge
    of the electron is -1).

    **Arguments**

    - **pids** : _numpy.ndarray_ or _list_
        - An array of numeric (float or integer) PDG ID values.
    - **error_on_unknown** : _bool_
        - Controls whether a `KeyError` is raised if an unknown PDG ID is
        encountered. If `False`, unknown PDG IDs will map to zero.

    **Returns**

    - _numpy.ndarray_
        - An array of charges as floats.
    """

    signs = np.sign(np.asarray(pids, dtype=float))
    abspids = np.abs(np.asarray(pids, dtype=int))
    orig_shape = abspids.shape
    abspids = abspids.reshape(-1)

    if error_on_unknown:
        charges = [PARTICLE_CHARGES[pid] for pid in abspids]
    else:
        charges = [PARTICLE_CHARGES.get(pid, 0.) for pid in abspids]

    return signs * np.asarray(charges, dtype=float).reshape(orig_shape)

def ischrgd(pids, ignored_pids=None):
    """Compute a boolean mask according to if the given PDG ID corresponds
    to a particle of non-zero charge.

    **Arguments**

    - **pids** : _numpy.ndarray_
        - An array of numeric (float or integer) PDG ID values.
    - **ignored_pids** : _numpy.ndarray_ or `None`
        - If not `None`, the PDG IDs in this array will not be considered
        charged, for instance to avoid counting quarks as charged particles.

    **Returns**

    - _numpy.ndarray_
        - A boolean mask corresponding to which particles are charged.
    """
    
    abspids = np.abs(np.asarray(pids, dtype=int))
    orig_shape = abspids.shape
    abspids = abspids.reshape(-1)

    if ignored_pids is None:
        charged = np.asarray([pid in CHARGED_PIDS for pid in abspids], dtype=bool)
    else:
        charged = np.asarray([(pid in CHARGED_PIDS) and (pid not in ignored_pids) 
                              for pid in abspids], dtype=bool)

    return charged.reshape(orig_shape)

LONG_METRIC = np.array([1.] + [-1.]*100)
def flat_metric(dim):
    """The Minkowski metric in `dim` spacetime dimensions in the mostly-minus
    convention.
    
    **Arguments**

    - **dim** : _int_
        - The number of spacetime dimensions (thought to be four in our 
        universe).

    **Returns**

    - _1-d numpy.ndarray_
        - A `dim`-length, one-dimensional (not matrix) array equal to 
        `[+1,-1,...,-1]`.
    """

    if dim <= 101:
        return LONG_METRIC[:dim]

    return np.asarray([1.] + [-1.]*(dim-1))



r"""Energy Flow Moments (EFMs) are tensors that can be computed in
$\mathcal O(M)$ where $M$ is the number of particles. They are useful for many
things, including providing a fast way of computing the $\beta=2$ EFPs, which
are the scalar contractions of products of EFMs.

The expression for a (normalized) hadronic EFM in terms of transverse momenta
$\{p_{Ti}\}$ and particle momenta $\{p_i^\mu\}$ is:
$$
\mathcal I^{\mu_1\cdots\mu_v} = 2^{v/2}\sum_{i=1}^Mz_in_i^{\mu_1}\cdots n_i^{\mu_v},
$$
where
$$
z_i=\frac{p_{Ti}}{\sum_jp_{Tj}},\quad\quad n_i^\mu=\frac{p_i^\mu}{p_{Ti}}.
$$
Note that for an EFM in an $e^+e^-$ context, transverse momenta are replaced
with energies.


Support for using EFMs to compute $\beta=2$ EFPs is built in to the `EFP` and
`EFPSet` classes using the classes and functions in this module. The `EFM` and
`EFMSet` classes can also be used on their own, as can the `efp2efms` function.
"""
from __future__ import absolute_import, division, print_function

from collections import OrderedDict
from operator import itemgetter
import sys

import numpy as np
from numpy.core.multiarray import c_einsum

from energyflow.algorithms import einsum
from energyflow.base import EFMBase
from energyflow.utils import flat_metric, timing
from energyflow.utils.graph_utils import *

__all__ = ['EFM', 'EFMSet', 'efp2efms']

###############################################################################
# EFM functions
###############################################################################

# allowed einsum symbols
I = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

def efp2efms(graph):
    """Translates an EFP formula, specified by its graph, to an expression
    involving EFMs. The input is a graph as a list of edges and the output is a
    tuple where the first argument is a string to be used with einsum and the
    second is a list of EFM signatures (the number of raised indices followed
    by the number of lowered indices).

    **Arguments**

    - **graph** : _list_ of _tuple_
        - The EFP graph given as a list of edges.

    **Returns**

    - (_str_, _list_ of _tuple_)
        - The einstring to be used with einsum for performing the contraction
        of EFMs followed by a list of the EFM specs. If `r` is the result of
        this function, and `efms` is a dictionary containing EFM tensors
        indexed by their signatures, then the value of the EFP is given as
        `np.einsum(r[0], *[efms[sig] for sig in r[1]])`.
    """

    # handle empty graph
    if len(graph) == 0:
        return '', [(0,0)]

    # build convenient data structure to hold graph information
    vds = get_valency_structure(graph)

    # dictionary to hold efm terms
    efms = {}

    # counter to store how to get fresh dummy indices
    ind = 0

    # iterate over vertices sorted by valency in decreasing order
    sorted_verts = sorted(valencies(graph).items(), key=itemgetter(1), reverse=True)
    for vert,valency in sorted_verts:

        # dict holding info for new efm term
        new_efm = {'upper_indices': '', 'lower_indices': ''}

        # iterate over neighboring vertices
        for neighbor,n_shared_edges in vds[vert].items():

            # if a new neighbor, assign fresh I
            if neighbor not in efms:
                new_I = I[ind:ind+n_shared_edges]
                ind += n_shared_edges
                new_efm['upper_indices'] += new_I

                # store I shared with that neighbor
                new_efm[neighbor] = new_I

            # if neighbor already has an efm factor, add already assigned indices to lower_indices
            else:
                new_efm['lower_indices'] += efms[neighbor][vert]

        # store new efm factor
        efms[vert] = new_efm

    einstr_list, efm_specs = [], []
    for vert,valency in sorted_verts:
        efm = efms[vert]

        # conventionally put uppered indices before lower indices
        einstr_list.append(efm['upper_indices'] + efm['lower_indices'])

        # add spec which is (nup, nlow) of efm
        efm_specs.append((len(efm['upper_indices']), len(efm['lower_indices'])))

    # return comma joined einstr and efm_specs
    return ','.join(einstr_list), efm_specs

###############################################################################
# EFM
###############################################################################

class EFM(EFMBase):

    """A class representing and computing a single EFM."""

    # EFM(nup, nlow=0, measure='hadrefm', beta=2, kappa=1, normed=None,
    #     coords=None, check_input=True)
    def __init__(self, nup, nlow=0, rl_from=None, subslice_from=None, **kwargs):
        r"""Since EFMs are fully symmetric tensors, they can be specified by
        just two integers: the number of raised and number of lowered indices
        that they carry. Thus we use a tuple of two ints as an EFM "spec" or
        signature throughout EnergyFlow. By convention the raised indices come
        before the lowered indices.

        Since a standalone `EFM` defines and holds a `Measure` instance, all
        `Measure` keywords are accepted. Note that `beta` is ignored as EFMs
        require $\beta=2$.

        **Arguments**

        - **nup** : _int_
            - The number of uppered indices of the EFM.
        - **nlow** : _int_
            - The number of lowered indices of the EFM.
        - **measure** : {`'hadrefm'`, `'eeefm'`}
            - The choice of measure. See [Measures](../measures) for additional
            info. Note that EFMs can only use the `'hadrefm'` and `'eeefm'`
            measures.
        - **beta** : _float_
            - The parameter $\beta$ appearing in the measure. Must be greater
            than zero.
        - **kappa** : {_float_, `'pf'`}
            - If a number, the energy weighting parameter $\kappa$. If `'pf'`,
            use $\kappa=v-1$ where $v$ is the valency of the vertex.
        - **normed** : _bool_
            - Controls normalization of the energies in the measure.
        - **coords** : {`'ptyphim'`, `'epxpypz'`, `None`}
            - Controls which coordinates are assumed for the input. See 
            [Measures](../measures) for additional info.
        - **check_input** : _bool_
            - Whether to check the type of the input each time or assume the
            first input type.
        """

        # initialize base class
        super(EFM, self).__init__(kwargs)

        # store inputs
        self._nup, self._nlow = nup, nlow
        self._rl_from = rl_from
        self._subslice_from = subslice_from

        # useful derived quantities
        self._v = self.nup + self.nlow
        self._spec = (self.nup, self.nlow)

        # construct by raising/lowering
        if self._rl_from is not None:

            # ensure valid raising/lowering
            if self.v != sum(self._rl_from):
                raise ValueError('cannot raise/lower among different valency EFMs')

            # determine einstr    
            diff = self.nup - self._rl_from[0]
            self._rl_diff = abs(diff)
            i_start, i_end = ((self._rl_from[0], self.nup) if diff > 0 else 
                              (self.nup, self._rl_from[0]))
            self.rl_einstr = ','.join([I[:self.v]] + list(I[i_start:i_end])) + '->' + I[:self.v]

            self._construct = self._rl_construct

        # construct by subslicing
        elif self._subslice_from is not None:

            # get number of subslices
            num_up_subslices = self._subslice_from[0] - self.nup
            num_low_subslices = self._subslice_from[1] - self.nlow
            
            # perform check
            if num_up_subslices < 0 or num_low_subslices < 0:
                m = 'invalid subslicing from {} to {}'.format(self.subslicing_from, self.spec)
                raise ValueError(m)

            # note that python 2 doesn't support pickling ellipsis
            if sys.version_info[0] > 2:
                self.subslice = tuple([0]*num_up_subslices + [Ellipsis] + [0]*num_low_subslices)
            else:
                self.subslice = tuple([0]*num_up_subslices + self.v*[slice(None)] + 
                                      [0]*num_low_subslices)

            self._pow2 = 2**(-(num_up_subslices + num_low_subslices)/2)
            self._construct = self._subslice_construct

        # construct directly
        else:
            self.raw_einstr = (','.join([I[0]] + [I[0] + I[i+1] for i in range(self.v)]) +
                               '->' + I[1:self.v+1])
            self.raw_einpath = ['einsum_path'] + [(0,1)]*self.v
            self._rl_diff = self.nlow
            self.rl_einstr = ','.join([I[:self.v]] + list(I[self.nup:self.v])) + '->' + I[:self.v]
            self._pow2 = 2**(self.v/2)
            self._construct = self._raw_construct

    #================
    # PRIVATE METHODS
    #================

    def _rl_construct(self, tensor):

        # fine to use pure c_einsum here as it's used anyway
        return c_einsum(self.rl_einstr, tensor, *[flat_metric(len(tensor))]*self._rl_diff)

    def _subslice_construct(self, tensor):
        return self._pow2 * tensor[self.subslice]

    def _raw_construct(self, zsnhats):
        zs, nhats = zsnhats
        M, dim = nhats.shape

        # if no lowering is needed
        if self.nlow == 0:
            return self._pow2 * einsum(self.raw_einstr, zs, *[nhats]*self.v, 
                                       optimize=self.raw_einpath)

        # lowering nhats first is better
        elif M*dim < dim**self.v:
            low_nhats = nhats * (flat_metric(dim)[np.newaxis])
            einsum_args = [nhats]*self.nup + [low_nhats]*self.nlow
            return self._pow2 * einsum(self.raw_einstr, zs, *einsum_args, 
                                       optimize=self.raw_einpath)

        # lowering EFM is better    
        else:
            tensor = einsum(self.raw_einstr, zs, *[nhats]*self.v, optimize=self.raw_einpath)
            return self._pow2 * self._rl_construct(tensor)

    #===============
    # PUBLIC METHODS
    #===============

    def compute(self, event=None, zs=None, nhats=None):
        """Evaluates the EFM on a single event. Note that `EFM` also is
        callable, in which case this method is invoked.

        **Arguments**

        - **event** : 2-d array_like or `fastjet.PseudoJet`
            - The event as an array of particles in the coordinates specified
            by `coords`.
        - **zs** : 1-d array_like
            - If present, `nhats` must also be present, and `zs` is used in place 
            of the energies of an event.
        - **nhats** : 2-d array like
            - If present, `zs` must also be present, and `nhats` is used in place
            of the scaled particle momenta.

        **Returns**

        - _numpy.ndarray_ of rank `v`
            - The values of the EFM tensor on the event. The raised indices
            are the first `nup` and the lowered indices are the last `nlow`.
        """

        return self._raw_construct(super(EFM, self).compute(event, zs, nhats))

    def batch_compute(self, events, n_jobs=None):
        """Evaluates the EFM on several events.

        **Arguments**

        - **events** : array_like or `fastjet.PseudoJet`
            - The events as an array of arrays of particles in coordinates
            matching those anticipated by `coords`.
        - **n_jobs** : _int_ or `None`
            - The number of worker processes to use. A value of `None` will
            use as many processes as there are CPUs on the machine.

        **Returns**

        - _numpy.ndarray_ of rank `v+1`
            - Array of EFM tensor values on the events.
        """

        return super(EFM, self).batch_compute(events, n_jobs)

    def set_timer(self):
        self.times = []
        self._construct = timing(self, self._construct)

    #===========
    # PROPERTIES
    #===========

    @property
    def nup(self):
        """The number of uppered indices on the EFM."""

        return self._nup

    @property
    def nlow(self):
        """The number of lowered indices on the EFM."""

        return self._nlow

    @property
    def spec(self):
        """The signature of the EFM as `(nup, nlow)`."""

        return self._spec

    @property
    def v(self):
        """The valency, or total number of indices, of the EFM."""

        return self._v

###############################################################################
# EFMSet
###############################################################################

class EFMSet(EFMBase):

    """A class for holding and efficiently constructing a collection of EFMs."""

    # EFMSet(efm_specs=None, vmax=None, measure='hadrefm', beta=2, kappa=1,
    #        normed=None, coords=None, check_input=True)
    def __init__(self, efm_specs=None, vmax=None, **kwargs):
        r"""An `EFMSet` can be initialized two ways (in order of precedence):

        1. **EFM Specs** - Pass in a list of EFM specs (`nup`, `nlow`).
        2. **Max Valency** - Specify a maximum valency and each EFM with up to
        that many indices will be constructed, with all indices raised.

        Since a standalone `EFMSet` defines and holds a `Measure` instance,
        all `Measure` keywords are accepted. Note that `beta` is ignored as
        EFMs require $\beta=2$.

        **Arguments**

        - **efm_specs** : {_list_, _tuple_, _set_} of _tuple_ or `None`
            - A collection of tuples of length two specifying which EFMs this
            object is to hold. Each spec is of the form `(nup, nlow)` where these
            are the number of upper and lower indices, respectively, that the EFM 
            is to have.
        - **vmax** : _int_
            - Only used if `efm_specs` is None. The maximum EFM valency to
            include in the `EFMSet`. Note that all EFMs will have `nlow=0`.
        - **measure** : {`'hadrefm'`, `'eeefm'`}
            - The choice of measure. See [Measures](../measures) for additional
            info. Note that EFMs can only use the `'hadrefm'` and `'eeefm'`
            measures.
        - **beta** : _float_
            - The parameter $\beta$ appearing in the measure. Must be greater
            than zero.
        - **kappa** : {_float_, `'pf'`}
            - If a number, the energy weighting parameter $\kappa$. If `'pf'`,
            use $\kappa=v-1$ where $v$ is the valency of the vertex.
        - **normed** : _bool_
            - Controls normalization of the energies in the measure.
        - **coords** : {`'ptyphim'`, `'epxpypz'`, `None`}
            - Controls which coordinates are assumed for the input. See 
            [Measures](../measures) for additional info.
        - **check_input** : _bool_
            - Whether to check the type of the input each time or assume the
            first input type.
        """

        hidden_subslicing = kwargs.pop('subslicing', False)

        # initialize base class
        super(EFMSet, self).__init__(kwargs)

        if efm_specs is None:
            if vmax is not None:
                vmin = 1 if self.normed else 0
                efm_specs = [(v,0) for v in range(vmin, vmax+1)]
            else:
                raise ValueError('efm_specs and vmax cannot both be None.')

        # get unique EFMs 
        self._unique_efms = frozenset(efm_specs)

        # setup EFMs based on whether we can subslice or not
        self.efms, self._args, self.rules = {}, {}, OrderedDict()
        if self.subslicing or hidden_subslicing:
            self._subslicing_setup()
        else:
            self._full_setup()

    #================
    # PRIVATE METHODS
    #================

    def _find_subslice(self, sig):
        """Determine if sig can be subsliced from the currently stored EFMs."""

        nup, nlow = sig
        bsigs = list(filter(lambda x: x[0] >= nup and x[1] >= nlow, self.efms))
        return min(bsigs, key=sum) if len(bsigs) else None

    def _find_minimum_rl(self, sig):
        v = sum(sig)
        vsigs = list(filter(lambda x: sum(x) == v, self.efms))
        return min(vsigs, key=lambda x: abs(sig[0]-x[0]))

    def _subslicing_setup(self):
        """Setup the rules for constructing the EFMs using the fact that
        setting an index to zero "pops" it off, which is referred to as the
        subclicing property. Typically, the EE measures have this property
        whereas the hadronic ones do not.
        """

        # ensure there is at least one EFM of each valency for rl purposes
        maxsig = max(self._unique_efms, key=sum) if len(self._unique_efms) else (0,0)
        self._unique_efms |= set((n,0) for n in range(1, sum(maxsig)+1))

        # sort EFMs to minimize raising/lowering operations
        # EFMs will be ordered first by decreasing v, then decreasing abs difference 
        # between nlow and nup, and then decreasing nup
        self._sorted_efms = sorted(self._unique_efms, key=itemgetter(0), reverse=True)
        self._sorted_efms.sort(key=lambda x: abs(x[0]-x[1]), reverse=True)
        self._sorted_efms.sort(key=sum, reverse=True)

        # take care of empty set
        if not len(self._sorted_efms):
            return

        # the first one must be raw constructed
        sig0 = self._sorted_efms[0]
        self.efms[sig0] = EFM(*sig0, no_measure=True)
        self._args[sig0] = 'r'
        self.rules[sig0] = 'constructing raw'

        for sig in self._sorted_efms[1:]:

            # determine if we can subslice
            big_spec = self._find_subslice(sig)
            if big_spec is not None:
                self.efms[sig] = EFM(*sig, subslice_from=big_spec, no_measure=True)
                self._args[sig] = big_spec
                self.rules[sig] = 'subslicing from {}'.format(big_spec)

            # find best raise/lower available
            else:
                rlsig = self._find_minimum_rl(sig)
                self.efms[sig] = EFM(*sig, rl_from=rlsig, no_measure=True)
                self._args[sig] = rlsig
                rl_n = abs(rlsig[0] - sig[0])
                self.rules[sig] = 'raising/lowering from {}, {}'.format(rlsig, rl_n)

    def _full_setup(self):
        """Setup the rules for constructing the EFMs without the assumption of any
        special properties.
        """

        # sort the EFMs first by increasing v and then by increasing nlow
        self._sorted_efms = sorted(self._unique_efms, key=itemgetter(1))
        self._sorted_efms.sort(key=sum)

        vprev, sigprev = None, None
        for sig in self._sorted_efms:
            v = sum(sig)

            # construct raw (all up) if this is a new valency
            if v != vprev:
                self.efms[sig] = EFM(*sig, no_measure=True)
                self._args[sig] = 'r'
                self.rules[sig] = 'constructing raw'

            # construct from lowering if we have a previous EFM with this v
            else:
                self.efms[sig] = EFM(*sig, rl_from=sigprev, no_measure=True)
                self._args[sig] = sigprev
                self.rules[sig] = 'lowering from {}'.format(sigprev)

            # update prevous values
            vprev, sigprev = v, sig

    #===============
    # PUBLIC METHODS
    #===============

    def compute(self, event=None, zs=None, nhats=None):
        """Evaluates the EFMs held by this `EFMSet` according to the
        predetermined strategy on a single event. Note that `EFMSet` also is
        callable, in which case this method is invoked.

        **Arguments**

        - **event** : 2-d array_like or `fastjet.PseudoJet`
            - The event as an array of particles in the coordinates specified
            by `coords`.
        - **zs** : 1-d array_like
            - If present, `nhats` must also be present, and `zs` is used in place 
            of the energies of an event.
        - **nhats** : 2-d array like
            - If present, `zs` must also be present, and `nhats` is used in place
            of the scaled particle momenta.

        **Returns**

        - _dict_ of _numpy.ndarray_ of rank `v`
            - A dictionary of EFM tensors indexed by their signatures.
        """

        zsnhats = super(EFMSet, self).compute(event, zs, nhats)

        efm_dict = {}
        for sig in self._sorted_efms:
            arg = self._args[sig]
            data_arg = zsnhats if arg == 'r' else efm_dict[arg]
            efm_dict[sig] = self.efms[sig]._construct(data_arg)

        return efm_dict

    def batch_compute(self, events, n_jobs=None):
        """Evaluates the EFMs held by the `EFMSet` on several events.

        **Arguments**

        - **events** : array_like or `fastjet.PseudoJet`
            - The events as an array of arrays of particles in coordinates
            matching those anticipated by `coords`.
        - **n_jobs** : _int_ or `None`
            - The number of worker processes to use. A value of `None` will
            use as many processes as there are CPUs on the machine.

        **Returns**

        - _numpy.ndarray_ of _dict_
            - Object array of dictionaries of EFM tensors indexed by their
            signatures.
        """

        return super(EFMSet, self).batch_compute(events, n_jobs)

    def set_timers(self):
        for efm in self.efms.values():
            efm.set_timer()

    def get_times(self):
        return {sig: np.asarray(efm.times) for sig,efm in self.efms.items()}

    #===========
    # PROPERTIES
    #===========

    @property
    def efms(self):
        """A dictionary of the `EFM` objects held by this `EFMSet` where the
        keys are the signatures of the EFM."""
        
        return self._efms

    @property
    def rules(self):
        """An ordered dictionary of the construction method used for each `EFM`
        where the order is the same as `sorted_efms`."""
        
        return self._rules

    @efms.setter
    def efms(self, value):
        self._efms = value

    @rules.setter
    def rules(self, value):
        self._rules = value



r"""## Z + Jets with Delphes Simulation

Datasets of QCD jets used for studying unfolding in [OmniFold: A Method to
Simultaneously Unfold All Observables](https://arxiv.org/abs/1911.09107). Four
different datasets are present:

- Herwig 7.1.5 with the default tune
- Pythia 8.243 with tune 21 (ATLAS A14 central tune with NNPDF2.3LO)
- Pythia 8.243 with tune 25 (ATLAS A14 variation 2+ of tune 21)
- Pythia 8.243 with tune 26 (ATLAS A14 variation 2- of tune 21)

$Z$ + jet events (with the $Z$ set to be stable) were generated for each of the
above generator/tune pairs with the $Z$ boson $\hat p_T^\text{min} > 150$ GeV
and $\sqrt{s}=14$ TeV. Events were then passed through the Delphes 3.4.2 fast
detector simulation of the CMS detector. Jets with radius parameter $R=0.4$
were found with the anti-kt algorithm at both particle level ("gen"), where all
non-neutrino, non-$Z$ particle are used, and detector level ("sim"), where
reconstructed energy flow objects (tracks, electromagnetic calorimeter cells,
and hadronic calorimeter cells) are used. Only jets with transverse momentum
greater than  are kept (note that sim jets have a simple jet energy correction
applied by Delphes). The hardest jet from events with a $Z$ boson with a final
transverse momentum of 200 GeV or greater are kept, yielding approximately 1.6
million jets at both gen-level and sim-level for each generator/tune pair.

Each zipped NumPy file consists of several arrays, the names of which begin
with either 'gen_' or 'sim_' depending on which set of jets they correspond
to. The name of each array ends in a key word indicating what it contains. With
the exception of 'gen_Zs' (which contains the $(p_T,\,y,\,\phi)$ of the final
$Z$ boson), there is both a gen and sim version of each array. The included
arrays are (listed by their key words):

- `'jets'` - The jet axis four vector, as $(p_T^\text{jet},\,y^\text{jet},\,
\phi^\text{jet},\,m^\text{jet})$ where $y^\text{jet}$ is the jet rapidity,
$\phi^\text{jet}$ is the jet azimuthal angle, and $m^\text{jet}$ is the jet
mass.
- `'particles'` - The rescaled and translated constituents of the jets as $(p_T/100,\,y-y^\text{jet},\,\phi-\phi^\text{jet},\,f_\text{PID})$ where
$f_\text{PID}$ is a small float corresponding to the PDG ID of the particle.
The PIDs are remapped according to $22\to0.0$, $211\to0.1$, $-211\to0.2$, $130
\to0.3$, $11\to0.4$, $-11\to0.5$, $13\to0.6$, $-13\to0.7$, $321\to0.8$, $-321
\to0.9$, $2212\to1.0$, $-2212\to1.1$, $2112\to1.2$, $-2112\to1.3$. Note that
ECAL cells are treated as photons (id 22) and HCAL cells are treated as 
$K_L^0$ (id 130). 
- `'mults'` - The constituent multiplicity of the jet.
- `'lhas'` - The Les Houches ($\beta=1/2$) angularity.
- `'widths'` - The jet width ($\beta=1 angularity).
- `'ang2s'` - The $\beta=2$ angularity (note that this is very similar to the jet
mass, but does not depend on particle masses).
- `'tau2s'` - The 2-subjettiness value with $\beta=1$.
- `'sdms'` - The groomed mass with Soft Drop parameters $z_\text{cut}=0.1$ and
$\beta=0$.
- `'zgs'` - The groomed momentum fraction (same Soft Drop parameters as above).

If you use this dataset, please cite the Zenodo record as well as the
corresponding paper:

[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.3548091.svg)](https://
doi.org/10.5281/zenodo.3548091) - Pythia/Herwig + Delphes samples

- A. Andreassen, P. T. Komiske, E. M. Metodiev, B. Nachman, J. Thaler, 
OmniFold: A Method to Simultaneously Unfold All Observables, [arXiv:1911.09107](https://arxiv.org/abs/1911.09107).
"""
from __future__ import absolute_import, division, print_function

import json
import os
import warnings

import numpy as np

from energyflow.utils import EF_DATA_DIR, ZENODO_URL_PATTERN
from energyflow.utils.data_utils import _get_filepath, _pad_events_axis1

__all__ = ['load']

DROPBOX_URL_PATTERN = 'https://www.dropbox.com/s/{}/{}?dl=1'
FILENAME_PATTERNS = {
    'pythia21': 'Pythia21_Zjet_pTZ-200GeV_{}.npz',
    'pythia25': 'Pythia25_Zjet_pTZ-200GeV_{}.npz',
    'pythia26': 'Pythia26_Zjet_pTZ-200GeV_{}.npz',
    'herwig': 'Herwig_Zjet_pTZ-200GeV_{}.npz'
}
DATASETS = frozenset(FILENAME_PATTERNS.keys())
KEYS = {
    # jets
    'jets': 'Jet four vectors, [pt, y, phi, m]',
    'particles': 'Jet constituents, [pt, y, phi, pid_float]',

    # Zs
    'Zs': 'Z momenta, [pt, y, phi]',

    # observables
    'ang2s': 'Jet angularity, beta = 2',
    'lhas': 'Les Houches Angularity, beta = 1/2',
    'mults': 'Jet constituent multiplicity',
    'sdms': 'Soft Dropped jet mass, zcut = 0.1, beta = 0',
    'tau2s': '2-subjettiness, beta = 1',
    'widths': 'Jet widths, beta = 1',
    'zgs': 'Groomed momentum fraction, zcut = 0.1, beta = 0',
}
NUM_FILES = 17
SOURCES = ['dropbox', 'zenodo']
ZENODO_RECORD = '3548091'

def load(dataset, num_data=100000, pad=False, cache_dir='~/.energyflow', source='zenodo', 
                  which='all', include_keys=None, exclude_keys=None):
    """Loads in the Z+jet Pythia/Herwig + Delphes datasets. Any file that is
    needed that has not been cached will be automatically downloaded.
    Downloaded files are cached for later use. Checksum verification is
    performed.

    **Arguments**

    - **datasets**: {`'Herwig'`, `'Pythia21'`, `'Pythia25'`, `'Pythia26'`}
        - The dataset (specified by which generator/tune was used to produce
        it) to load. Note that this argument is not sensitive to
        capitalization.
    - **num_data**: _int_
        - The number of events to read in. A value of `-1` means to load all
        available events.
    - **pad**: _bool_
        - Whether to pad the particles with zeros in order to form contiguous
        arrays.
    - **cache_dir**: _str_
        - Path to the directory where the dataset files should be stored.
    - **source**: {`'dropbox'`, `'zenodo'`}
        - Which location to obtain the files from.
    - **which**: {`'gen'`, `'sim'`, `'all'`}
        - Which type(s) of events to read in. Each dataset has corresponding
        generated events at particle-level and simulated events at
        detector-level.
    - **include_keys**: _list_ or_tuple_ of _str_, or `None`
        - If not `None`, load these keys from the dataset files. A value of
        `None` uses all available keys (the `KEYS` global variable of this
        module contains the available keys as keys of the dictionary and
        brief descriptions as values). Note that keys do not have 'sim' or
        'gen' prepended to the front yet.
    - **exclude_keys**: _list_ or _tuple_ or _str_, or `None`
        - Any keys to exclude from loading. Most useful when a small number of
        keys should be excluded from the default set of keys.

    **Returns**

    - _dict_ of _numpy.ndarray_
        - A dictionary of the jet, particle, and observable arrays for the
        specified dataset.
    """

    # load info from JSON file
    if 'INFO' not in globals():
        global INFO
        with open(os.path.join(EF_DATA_DIR, 'ZjetsDelphes.json'), 'r') as f:
            INFO = json.load(f)

    # check that options are valid
    dataset_low = dataset.lower()
    if dataset_low not in DATASETS:
        raise ValueError("Invalid dataset '{}'".format(dataset))
    if source not in SOURCES:
        raise ValueError("Invalud source '{}'".format(source))

    # handle selecting keys
    keys = set(KEYS.keys()) if include_keys is None else set(include_keys)
    keys -= set() if exclude_keys is None else set(exclude_keys)
    for key in keys:
        if key not in KEYS:
            raise ValueError("Unrecognized key '{}'".format(key))

    # create dictionray to store values to be returned
    levels = ['gen', 'sim'] if which == 'all' else [which.lower()]
    for level in levels:
        if level != 'gen' and level != 'sim':
            raise ValueError("Unrecognized specification '{}' ".format(level) +
                             "for argument 'which', allowed options are 'all', 'gen', 'sim'")
    vals = {'{}_{}'.format(level, key) : [] for level in levels for key in keys}
    if 'sim_Zs' in vals:
        del vals['sim_Zs']

    # get filenames
    filenames = [FILENAME_PATTERNS[dataset_low].format(i) for i in range(NUM_FILES)]

    # get urls
    if source == 'dropbox':
        db_link_hashes = INFO['dropbox_link_hashes'][dataset_low]
        urls = [DROPBOX_URL_PATTERN.format(dl, fn) for dl,fn in zip(db_link_hashes, filenames)]
    elif source == 'zenodo':
        urls = [ZENODO_URL_PATTERN.format(ZENODO_RECORD, fn) for fn in filenames]

    # get hashes
    hashes = INFO['hashes'][dataset_low]['sha256']

    n = 0
    subdir = os.path.join('datasets', 'ZjetsDelphes')
    for filename, url, h in zip(filenames, urls, hashes):

        # check if we have enough events
        if n >= num_data and num_data != -1:
            break

        # load file
        f = np.load(_get_filepath(filename, url, cache_dir, cache_subdir=subdir, file_hash=h))

        # add relevant arrays to vals
        for i,(key,val) in enumerate(vals.items()):

            if 'particles' not in key or pad:
                val.append(f[key])
            else:
                val.append([np.array(ps[ps[:,0] > 0]) for ps in f[key]])

            # increment number of events
            if i == 0:
                n += len(val[-1])

        f.close()

    # warn if we don't have enough events
    if num_data > n:
        warnings.warn('Only have {} events when {} were requested'.format(n, num_data))

    # concatenate arrays
    s = slice(0, num_data if num_data > -1 else None)
    for key,val in vals.items():

        if 'particles' not in key or not pad:
            vals[key] = np.concatenate(val, axis=0)[s]
        else:
            max_len_axis1 = max([X.shape[1] for X in val])
            vals[key] = np.vstack([_pad_events_axis1(x, max_len_axis1) for x in val])[s]

    return vals



r"""## Quark and Gluon Nsubs

A dataset consisting of 45 $N$-subjettiness observables for 100k quark and 
gluon jets generated with Pythia 8.230. Following [1704.08249](https:
//arxiv.org/abs/1704.08249), the observables are in the following order:

$$
\{\tau_1^{(\beta=0.5)},\tau_1^{(\beta=1.0)},\tau_1^{(\beta=2.0)},
\tau_2^{(\beta=0.5)},\tau_2^{(\beta=1.0)},\tau_2^{(\beta=2.0)},
\ldots,
\tau_{15}^{(\beta=0.5)},\tau_{15}^{(\beta=1.0)},\tau_{15}^{(\beta=2.0)}\}.
$$

The dataset contains two members: `'X'` which is a numpy array of the nsubs
that has shape `(100000,45)` and `'y'` which is a numpy array of quark/gluon 
labels (quark=`1` and gluon=`0`).
"""
from __future__ import absolute_import, division, print_function

import numpy as np

from energyflow.utils.data_utils import _get_filepath

__all__ = ['load']

def load(num_data=-1, cache_dir='~/.energyflow'):
    """Loads the dataset. The first time this is called, it will automatically
    download the dataset. Future calls will attempt to use the cached dataset 
    prior to redownloading.

    **Arguments**

    - **num_data** : _int_
        - The number of events to return. A value of `-1` means read in all events.
    - **cache_dir** : _str_
        - The directory where to store/look for the file.

    **Returns**

    - _3-d numpy.ndarray_, _1-d numpy.ndarray_
        - The `X` and `y` components of the dataset as specified above.
    """

    fpath = _get_filepath('QG_nsubs.npz', 
                      url='https://www.dropbox.com/s/y1l6avj5yj7jn9t/QG_nsubs.npz?dl=1',
                      file_hash='a99f771147af9b207356c990430cfeba6b6aa96fe5cff8263450ff3a31ab0997',
                      cache_dir=cache_dir)

    f = np.load(fpath)
    X, y = f['X'], f['y']
    f.close()

    if num_data > -1:
        X, y = X[:num_data], y[:num_data]

    return X, y
    